/*
 * FFplay : Simple Media Player based on the ffmpeg libraries
 * Copyright (c) 2003 Fabrice Bellard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <pspkernel.h>
#include <pspdebug.h>
#include <psppower.h>
#include <pspthreadman.h>
#include <pspiofilemgr.h>
#include <pspaudio.h>
#include <pspaudiolib.h>
#include <pspctrl.h>
#include <pspgu.h>
#include <sys/types.h>
#include <sys/unistd.h>
#include "PspLib/pg.h"
#include "myscreen.h"
#include "mycolor.h"
#include "mydrawbmp.h"
#include "mydrawtext.h"
#include "mytext.h"
#include "mystring.h"
#include "HanLib/han_lib.h"
#include "MY_ttf.h"
#include "myjpeg.h"
#include "myevent.h"
#include "mytimer.h"
#include "myaudio.h"
#include "avformat.h"											 

//#define JONNY_YUV2RGB				   

//#ifdef JONNY_YUV2RGB
#include "mem64.h"                      // [jonny]
#include "me.h"                         // [jonny]
#include "me_csc.h"                     // [jonny]
//volatile struct me_struct *p_me_struct; // [jonny]
//#endif

#ifdef CONFIG_WIN32
#undef main /* We don't want SDL to override our main() */
#endif

#if defined(__linux__)
#define HAVE_X11
#endif

#ifdef HAVE_X11
#include <X11/Xlib.h>
#endif

//#define DEBUG_SYNC

#define MAX_VIDEOQ_SIZE (4 * 256 * 1024)
#define MAX_AUDIOQ_SIZE (4 * 16 * 1024)
//#define MAX_AUDIOQ_SIZE (16 * 1024)
#define MAX_SUBTITLEQ_SIZE (5 * 16 * 1024)

/* SDL audio buffer size, in samples. Should be small to have precise
   A/V sync as SDL does not have hardware buffer fullness info. */
//#define SDL_AUDIO_BUFFER_SIZE 1024
#define SDL_AUDIO_BUFFER_SIZE 1024

/* no AV sync correction is done if below the AV sync threshold */
#define AV_SYNC_THRESHOLD 0.01
/* no AV correction is done if too big error */
#define AV_NOSYNC_THRESHOLD 10.0

/* maximum audio speed change to get correct sync */
#define SAMPLE_CORRECTION_PERCENT_MAX 10

/* we use about AUDIO_DIFF_AVG_NB A-V differences to make the average */
#define AUDIO_DIFF_AVG_NB   20

/* NOTE: the size must be big enough to compensate the hardware audio buffersize size */
#define SAMPLE_ARRAY_SIZE (2*65536)

struct Vertex
{
	float u,v;
	unsigned int color;
	float x,y,z;
};

typedef struct YUV_Overlay {
	int w, h;				/* Read-only */
	u16 pitches[3];
	u8 *raw_buf;
	u8 *planes[3];
} YUV_Overlay;

typedef struct PacketQueue {
    AVPacketList *first_pkt, *last_pkt;
    int nb_packets;
    int size;
    int abort_request;
	SceUID mutex;
} PacketQueue;

#define VIDEO_PICTURE_QUEUE_SIZE 1
#define SUBPICTURE_QUEUE_SIZE 4

typedef struct VideoPicture {
    double pts; /* presentation time stamp for this picture */
#ifdef JONNY_YUV2RGB
	AVFrame *bmp;
	uint8_t *buffer;
#else
	YUV_Overlay *bmp;
#endif
    int width, height; /* source height & width */
    int allocated;
} VideoPicture;

typedef struct SubPicture {
    double pts; /* presentation time stamp for this picture */
    AVSubtitle sub;
} SubPicture;

enum {
    AV_SYNC_AUDIO_MASTER, /* default choice */
    AV_SYNC_VIDEO_MASTER,
    AV_SYNC_EXTERNAL_CLOCK, /* synchronize to an external clock */
};

struct AVResampleContext;

struct ReSampleContext {
    struct AVResampleContext *resample_context;
    short *temp[2];
    int temp_len;
    float ratio;
    /* channel convert */
    int input_channels, output_channels, filter_channels;
};

typedef struct VideoState {
	SceUID parse_tid;
	SceUID video_tid;
    AVInputFormat *iformat;
    int no_background;
    int abort_request;
    int paused;
    int last_paused;
    int seek_req;
	int seek_flags;
    int64_t seek_pos;
    AVFormatContext *ic;
    int dtg_active_format;

    int audio_stream;
    
    int av_sync_type;
    double external_clock; /* external clock base */
    int64_t external_clock_time;
    
    double audio_clock;
    double audio_diff_cum; /* used for AV difference average computation */
    double audio_diff_avg_coef;
    double audio_diff_threshold;
    int audio_diff_avg_count;
    AVStream *audio_st;
    PacketQueue audioq;
    int audio_hw_buf_size;
    unsigned int audio_buf_size; /* in bytes */
    int audio_buf_index; /* in bytes */
    AVPacket audio_pkt;
    uint8_t *audio_pkt_data;
    int audio_pkt_size;
    
    int show_audio; /* if true, display audio samples */
    int16_t sample_array[SAMPLE_ARRAY_SIZE];
    int sample_array_index;
    int last_i_start;
    
    double frame_timer;
    double frame_last_pts;
    double frame_last_delay;
    double video_clock;
    int video_stream;
    AVStream *video_st;
    PacketQueue videoq;
    double video_current_pts; /* current displayed pts (different from
                                 video_clock if frame fifos are used) */
    int64_t video_current_pts_time; /* time at which we updated
                                       video_current_pts - used to
                                       have running video pts */
    VideoPicture pictq[VIDEO_PICTURE_QUEUE_SIZE];
    int pictq_size, pictq_rindex, pictq_windex;
	SceUID pictq_mutex;
    
	SceUID video_decoder_mutex;
    SceUID audio_decoder_mutex;
	SceUID audio_buffer_mutex;
    //    QETimer *video_timer;
    char filename[1024];
    int width, height, xleft, ytop;
	int audio_resample;
    ReSampleContext *resample; /* for audio resampling */
    /* samples output by the codec. we reserve more space for avsync
       compensation */
//    uint8_t __attribute__((aligned(16))) audio_buf[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2]; 
    uint8_t audio_buf[AVCODEC_MAX_AUDIO_FRAME_SIZE * 2];
} VideoState;

typedef enum
{
	MEDIA_VIDEO,
	MEDIA_MUSIC
} media_type;

my_bmp_type *background;

typedef struct
{
	char filename[MAXPATHLEN];
	int ptsize;
	TTF_Font *font;
	int renderstyle;
	int rendersolid;
	int fontheight;
} font_info;

font_info my_font;
#define DEFAULT_FONTNAME "font/default.ttf"
#define DEFAULT_PTSIZE	13
#define TITLE_PTSIZE 15
#define BIG_PTSIZE 20
extern int ufEuc_kr2Unicode (unsigned int * dest, unsigned char* src);

void show_help(void);
static int audio_write_get_buf_size(VideoState *is);
void sdl_audio_callback(short* buffer, unsigned long length);
int get_prev_file(media_type type);
int get_next_file(media_type type);
static void stream_pause(VideoState *is);
static void stream_seek(VideoState *is, int64_t pos, int rel);
void do_exit(void);
static VideoState *stream_open(const char *filename, AVInputFormat *iformat);
static double get_master_clock(VideoState *is);
void show_info_box(const char *msg);
void msg_box(const char *msg);
void clear_screen(void);
void reset_background(void);
void exit_program(void);
void ascii2unicode(u8 *src, u16* dest);
static void video_refresh_timer(void *opaque);

/* options specified by the user */
static AVInputFormat *file_iformat;
static AVImageFormat *image_format;
static const char *input_filename;
static char FileName[MAXPATHLEN];
static int audio_disable;
static int video_disable;
static int display_disable;
static int show_status;
static int av_sync_type = AV_SYNC_AUDIO_MASTER;
static int64_t start_time = AV_NOPTS_VALUE;
static int debug = 0;
static int debug_mv = 0;
static int step = 0;
static int thread_count = 1;
static int workaround_bugs = 1;

static int fast = 0;
static int genpts = 0;
static int lowres = 0;
static int idct = FF_IDCT_AUTO;
static enum AVDiscard skip_frame= AVDISCARD_DEFAULT;
static enum AVDiscard skip_idct= AVDISCARD_DEFAULT;
static enum AVDiscard skip_loop_filter= AVDISCARD_DEFAULT;
static int error_resilience = FF_ER_CAREFUL;
static int error_concealment = 3;

/* current context */
static VideoState *cur_stream;
static int64_t audio_callback_time;

int done = 0;

static int need_alloc = 0;
//static double next_refresh_time;
//static int need_timer_check;
static int disp_w, disp_h;
static int play_finished = 0;
static char saved_filename[MAXPATHLEN];
static double saved_pos;
static int return_from_sleep = 0;
static int catched_error = 0;
static int skip_frames = 0;
static int display_clock = 0;
static char clock_text[256] = {"clock : 266"};

u16 unicode_text[2048];

#define MAX_MEDIA_FILES 100

static char media_dir_name[MAXPATHLEN];

static SceIoDirent media_file_list[MAX_MEDIA_FILES];
static int media_file_num;
static int start_idx;
static int last_start_idx;
static int media_idx;
static int last_media_idx;
static int list_end;
static unsigned char has_scrolbar = 0;
u8 full_screen = 0;
int language = LANGUAGE_KOREAN;

int use_ttf_font = 1;
int ttf_failed = 0;
int refresh_timer_id = -1;
int repeat = 0;

typedef enum
{
	NORMAL_MODE,
	FIT_MODE,
	FULL_SCREEN
} screen_mode;
screen_mode mode = NORMAL_MODE;
void set_disp_mode(VideoState *is, screen_mode mode);

PSP_MODULE_INFO("PSP Media Player", 0x1000, 1, 1);
PSP_MAIN_THREAD_ATTR(0);
PSP_MAIN_THREAD_STACK_SIZE_KB(30); /* smaller stack for kernel thread */

//__attribute__((constructor)) void stdoutInit() 
//{ 
//   pspKernelSetKernelPC(); 
//   pspDebugInstallStderrHandler(pspDebugScreenPrintData); 
//}

void MyExceptionHandler(PspDebugRegBlock * regs)
{
    /* I always felt BSODs were more interesting that white on black */

    pspDebugScreenSetBackColor(0x00FF0000);
    pspDebugScreenSetTextColor(0xFFFFFFFF);

    pspDebugScreenSetXY(0,17);
    pspDebugScreenPrintf("\nI regret to inform you your psp has just crashed\n");
    pspDebugScreenPrintf("Exception Details:\n");
    pspDebugDumpException(regs);
    pspDebugScreenSetBackColor(0x00000000);

}

/* Exit callback */
int exit_callback(int arg1, int arg2, void *common)
{
	done = 1;
	return 0;
}

void PowerCallback (int unknown, int pwrflags)
{
	if (pwrflags & PSP_POWER_CB_HOLD_SWITCH)
	{
	}
	
	if (pwrflags & PSP_POWER_CB_POWER_SWITCH)
	{
        if (cur_stream) {
#ifndef JONNY_YUV2RGB
            saved_pos = get_master_clock(cur_stream);
			strcpy(saved_filename, cur_stream->filename);
#else
			done = 1; // not working ... why?
#endif
        }
		else
		{
#ifndef JONNY_YUV2RGB
			saved_pos = -1;
#else
			done = 1; // not working ... why?
#endif
		}
	}
	else if (pwrflags & PSP_POWER_CB_RESUME_COMPLETE) 
	{
		if (saved_pos >= 0)
		{
#ifndef JONNY_YUV2RGB
			return_from_sleep = 1;
#else
#endif
		}
	}

	if (pwrflags & PSP_POWER_CB_BATTERY_LOW){
	}

	int cbid;
	cbid = sceKernelCreateCallback ("Power Callback", (SceKernelCallbackFunction)PowerCallback, NULL);
	scePowerRegisterCallback       (0, cbid);
}

/* Callback thread */
int CallbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);
	cbid = sceKernelCreateCallback ("Power Callback", (SceKernelCallbackFunction)PowerCallback, NULL);
	scePowerRegisterCallback       (0, cbid);
	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, THREAD_ATTR_USER, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

void print_error(const char *filename, int err)
{
	char msg[2048];

    switch(err) {
    case AVERROR_NUMEXPECTED:
        snprintf(msg, 2048, "%s: Incorrect image filename syntax.\n"
                "Use '%%d' to specify the image number:\n"
                "  for img1.jpg, img2.jpg, ..., use 'img%%d.jpg';\n"
                "  for img001.jpg, img002.jpg, ..., use 'img%%03d.jpg'.\n", 
                filename);
		msg_box((const char *)msg);
        break;
    case AVERROR_INVALIDDATA:
        snprintf(msg, 2048, "%s: Error while parsing header\n", filename);
		msg_box((const char *)msg);
        break;
    case AVERROR_NOFMT:
        snprintf(msg, 2048, "%s: Unknown format\n", filename);
		msg_box((const char *)msg);
        break;
    default:
        snprintf(msg, 2048, "%s: Error while opening file\n", filename);
		msg_box((const char *)msg);
        break;
    }
}

static unsigned int __attribute__((aligned(16))) y_clut[256];
static unsigned int __attribute__((aligned(16))) add_u_clut[256];
static unsigned int __attribute__((aligned(16))) sub_u_clut[256];
static unsigned int __attribute__((aligned(16))) add_v_clut[256];
static unsigned int __attribute__((aligned(16))) sub_v_clut[256];

void init_palettes()
{
	int i;
	u8 r, g, b;

	/* palette #0 : Y values */
	/* it seems like y values are in [0..255] not [16..235] so we don't need to do any calc. on y values */
	for (i = 0; i < 256; i++)
	{
		if (i < 16 || i > 235)
			y_clut[i] = (0xFF000000 | (i << 16) | (i << 8) | (i << 0));
		else
		{
			u8 Y;
			Y = (i - 16) * 1.164;
			y_clut[i] = (0xFF000000 | (Y << 16) | (Y << 8) | (Y << 0));
		}
	}

	/* palette #1 : additive U values */
	r = 0;
	for (i = 0; i < 256; i++)
	{
		if (i < 128)
//			g = (128 - i) * 0.338;
			g = (128 - i) * 0.391;
		else
			g = 0;

		if (i > 128)
//			b = (i - 128) * 1.732;
			b = (i - 128) * 2.018;
		else
			b = 0;

		add_u_clut[i] = (0xFF000000 | (b << 16) | (g << 8) | (r << 0));
	}

	/* palette #2 : subtractive U values */
	r = 0;
	for (i = 0; i < 256; i++)
	{
		if (i > 128)
//			g = (i - 128) * 0.338;
			g = (i - 128) * 0.391;
		else
			g = 0;

		if (i < 128)
//			b = (128 - i) * 1.732;
			b = (128 - i) * 2.018;
		else
			b = 0;

		sub_u_clut[i] = (0xFF000000 | (b << 16) | (g << 8) | (r << 0));
	}

	/* palette #3 : additive V values */
	b = 0;
	for (i = 0; i < 256; i++)
	{
		if (i < 128)
//			g = (128 - i) * 0.698;
			g = (128 - i) * 0.813;
		else
			g = 0;

		if (i > 128)
//			r = (i - 128) * 1.370;
			r = (i - 128) * 1.596;
		else
			r = 0;

		add_v_clut[i] = (0xFF000000 | (b << 16) | (g << 8) | (r << 0));
	}

	/* pallete #4 : subtractive V values */
	b = 0;
	for (i = 0; i < 256; i++)
	{
		if (i > 128)
//			g = (i - 128) * 0.698;
			g = (i - 128) * 0.813;
		else
			g = 0;

		if (i < 128)
//			r = (128 - i) * 1.370;
			r = (128 - i) * 1.596;
		else
			r = 0;

		sub_v_clut[i] = (0xFF000000 | (b << 16) | (g << 8) | (r << 0));
	}
}

YUV_Overlay *create_overlay_surface(int width, int height)
{
	YUV_Overlay *ov;
	int surfwidth;

	ov = (YUV_Overlay *)av_malloc(sizeof(YUV_Overlay));
	memset(ov, 0, sizeof(YUV_Overlay));
	if (!ov)
	{
		return NULL;
	}
	ov->w = width;
	ov->h = height;

	/* I know this code is sucks */
	if (width < 512)
		surfwidth = 512;
	else
		surfwidth = 1024;

	ov->pitches[0] = surfwidth;
	ov->pitches[1] = surfwidth/2;
	ov->pitches[2] = surfwidth/2;

	ov->planes[0] = (u8 *)av_malloc(surfwidth * height);
	if (!ov->planes[0])
	{
		av_free(ov);
		return NULL;
	}
	ov->planes[1] = (u8 *)av_malloc(surfwidth * height / 4);
	if (!ov->planes[1])
	{
		av_free(ov->planes[0]);
		av_free(ov);
		return NULL;
	}
	ov->planes[2] = (u8 *)av_malloc(surfwidth * height / 4);
	if (!ov->planes[2])
	{
		av_free(ov->planes[1]);
		av_free(ov->planes[0]);
		av_free(ov);
		return NULL;
	}
	return ov;
}

void free_overlay_surface(YUV_Overlay *ov)
{
	if (ov)
	{
		av_free(ov->planes[2]);
		av_free(ov->planes[1]);
		av_free(ov->planes[0]);
		av_free(ov);
	}
}

#define SLICE_SIZE 64 // Change this to experiment with different page-cache sizes
#define Y_PLANE 0
#define ADD_UV_PLANE 1
#define SUB_UV_PLANE 2

struct BlendState
{
	int op;
	int src;
	int dest;
	unsigned int srcfix;
	unsigned int destfix;
};

struct BlendState states[] =
{
	{GU_ADD, GU_FIX, GU_FIX, 0xffffffff, 0xffffffff},
	{GU_REVERSE_SUBTRACT, GU_FIX, GU_FIX, 0xffffffff, 0xffffffff},
};

void blit_y_plane(u8 *tex, int src_w,int src_h,int src_pitch,int dst_x, int dst_y, int dst_w,int dst_h)
{
	unsigned int j,cx,cy;
	struct Vertex* vertices;
  	unsigned int* clut = (unsigned int*)(((unsigned int)y_clut)|0x40000000);

	cx = dst_x;
	cy = dst_y;

	sceKernelDcacheWritebackAll();

	sceGuScissor(cx,cy,dst_w,dst_h);

	sceGuBlendFunc( GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0 );

	/* Upload the palette to color lookup table */
	sceGuClutMode(GU_PSM_8888,0,0xff,0); // 32-bit palette
	sceGuClutLoad(((256)/8),clut); // upload 32*8 entries (256)

	sceGuTexMode(GU_PSM_T8,0,0,0);
	sceGuTexImage(0,src_pitch,src_pitch,src_pitch,tex);
	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
	sceGuTexFilter(GU_LINEAR,GU_LINEAR);
	sceGuTexScale(1.0f/src_pitch,1.0f/src_pitch); // scale UVs to 0..1
	sceGuTexOffset(0,0);
	sceGuAmbientColor(0xffffffff);
	sceGuColor(0xffffffff);
  
	for (j = 0; (j+SLICE_SIZE) <= src_w; j = j+SLICE_SIZE) 
	{
    	vertices = (struct Vertex*)sceGuGetMemory(2 * sizeof(struct Vertex));
    
	    vertices[0].u = j; vertices[0].v = 0;
    
    	vertices[0].x = cx+j*dst_w/src_w; vertices[0].y = cy+0; vertices[0].z = 0;
    
	    vertices[1].u = j+SLICE_SIZE; vertices[1].v = src_h;
    
    	vertices[1].x = cx+(j+SLICE_SIZE)*dst_w/src_w; vertices[1].y = cy+dst_h; vertices[1].z = 0;
    
		sceGuDrawArray(GU_SPRITES,GU_TEXTURE_32BITF|GU_COLOR_8888|GU_VERTEX_32BITF|GU_TRANSFORM_2D,2,0,vertices);
	}

	if (j<src_w)
	{
    	vertices = (struct Vertex*)sceGuGetMemory(2 * sizeof(struct Vertex));
    
	    vertices[0].u = j; vertices[0].v = 0;
    
    	vertices[0].x = cx+j*dst_w/src_w; vertices[0].y = cy+0; vertices[0].z = 0;
	    vertices[1].u = src_w; vertices[1].v = src_h;
    
    	vertices[1].x = cx+dst_w; vertices[1].y = cy+dst_h; vertices[1].z = 0;
    
		sceGuDrawArray(GU_SPRITES,GU_TEXTURE_32BITF|GU_COLOR_8888|GU_VERTEX_32BITF|GU_TRANSFORM_2D,2,0,vertices);
	}
}

void blit_uv_plane(u8 *tex, int src_w,int src_h,int src_pitch,int dst_x, int dst_y, int dst_w,int dst_h, int mode, unsigned int *table)
{
	unsigned int j,cx,cy;
	struct Vertex* vertices;
  	unsigned int* clut = (unsigned int*)(((unsigned int)table)|0x40000000);
  	
	cx = dst_x;
	cy = dst_y;

	sceKernelDcacheWritebackAll();

	sceGuScissor(cx,cy,dst_w,dst_h);

	sceGuBlendFunc( states[mode].op, states[mode].src, states[mode].dest, states[mode].srcfix, states[mode].destfix );

	/* Upload the palette to color lookup table */
	sceGuClutMode(GU_PSM_8888,0,0xff,0); // 32-bit palette
	sceGuClutLoad(((256)/8),clut); // upload 32*8 entries (256)

	sceGuTexMode(GU_PSM_T8,0,0,0);
	sceGuTexImage(0,src_pitch,src_pitch,src_pitch,tex);
	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
	sceGuTexFilter(GU_LINEAR,GU_LINEAR);
	sceGuTexScale(1.0f/src_pitch,1.0f/src_pitch); // scale UVs to 0..1
	sceGuTexOffset(0,0);
	sceGuAmbientColor(0xffffffff);
	sceGuColor(0xffffffff);
  
	for (j = 0; (j+SLICE_SIZE) <= src_w; j = j+SLICE_SIZE) 
	{
    	vertices = (struct Vertex*)sceGuGetMemory(2 * sizeof(struct Vertex));
    
	    vertices[0].u = j; vertices[0].v = 0;
    
    	vertices[0].x = cx+j*dst_w/src_w; vertices[0].y = cy+0; vertices[0].z = 0;
    
	    vertices[1].u = j+SLICE_SIZE; vertices[1].v = src_h;
    
    	vertices[1].x = cx+(j+SLICE_SIZE)*dst_w/src_w; vertices[1].y = cy+dst_h; vertices[1].z = 0;
    
		sceGuDrawArray(GU_SPRITES,GU_TEXTURE_32BITF|GU_COLOR_8888|GU_VERTEX_32BITF|GU_TRANSFORM_2D,2,0,vertices);
	}

	if (j<src_w)
	{
    	vertices = (struct Vertex*)sceGuGetMemory(2 * sizeof(struct Vertex));
    
	    vertices[0].u = j; vertices[0].v = 0;
    
    	vertices[0].x = cx+j*dst_w/src_w; vertices[0].y = cy+0; vertices[0].z = 0;
	    vertices[1].u = src_w; vertices[1].v = src_h;
    
    	vertices[1].x = cx+dst_w; vertices[1].y = cy+dst_h; vertices[1].z = 0;
    
		sceGuDrawArray(GU_SPRITES,GU_TEXTURE_32BITF|GU_COLOR_8888|GU_VERTEX_32BITF|GU_TRANSFORM_2D,2,0,vertices);
	}
}

/* packet queue handling */
static void packet_queue_init(PacketQueue *q)
{
    memset(q, 0, sizeof(PacketQueue));
	q->mutex = sceKernelCreateSema("", 0, 1, 1, 0);
}

static void packet_queue_flush(PacketQueue *q)
{
    AVPacketList *pkt, *pkt1;

	sceKernelWaitSema(q->mutex, 1, 0);
    for(pkt = q->first_pkt; pkt != NULL; pkt = pkt1) {
        pkt1 = pkt->next;
        av_free_packet(&pkt->pkt);
		av_freep(&pkt);
    }
    q->last_pkt = NULL;
    q->first_pkt = NULL;
    q->nb_packets = 0;
    q->size = 0;
	sceKernelSignalSema(q->mutex, 1);
}

static void packet_queue_end(PacketQueue *q)
{
    packet_queue_flush(q);
	sceKernelDeleteSema(q->mutex);
}

static int packet_queue_put(PacketQueue *q, AVPacket *pkt)
{
    AVPacketList *pkt1;

	sceKernelWaitSema(q->mutex, 1, 0);
    /* duplicate the packet */
    if (av_dup_packet(pkt) < 0)
	{
		sceKernelSignalSema(q->mutex, 1);
        return -1;
	}
    
    pkt1 = av_malloc(sizeof(AVPacketList));
    if (!pkt1)
	{
		sceKernelSignalSema(q->mutex, 1);
        return -1;
	}
    pkt1->pkt = *pkt;
    pkt1->next = NULL;

    if (!q->last_pkt)

        q->first_pkt = pkt1;
    else
        q->last_pkt->next = pkt1;
    q->last_pkt = pkt1;
    q->nb_packets++;
    q->size += pkt1->pkt.size;
    /* XXX: should duplicate packet data in DV case */

	sceKernelSignalSema(q->mutex, 1);
    return 0;
}

static void packet_queue_abort(PacketQueue *q)
{
	sceKernelWaitSema(q->mutex, 1, 0);

    q->abort_request = 1;
    
	sceKernelSignalSema(q->mutex, 1);
}

/* return < 0 if aborted, 0 if no packet and > 0 if packet.  */
static int packet_queue_get(PacketQueue *q, AVPacket *pkt, int block)
{
    AVPacketList *pkt1;
	int ret;

    for(;;) {
		sceKernelWaitSema(q->mutex, 1, 0);
        if (q->abort_request) {
            ret = -1;
			sceKernelSignalSema(q->mutex, 1);
            break;
        }
        pkt1 = q->first_pkt;
        if (pkt1) {
            q->first_pkt = pkt1->next;
            if (!q->first_pkt)
                q->last_pkt = NULL;
            q->nb_packets--;
            q->size -= pkt1->pkt.size;
            *pkt = pkt1->pkt;
            av_free(pkt1);
            ret = 1;
			sceKernelSignalSema(q->mutex, 1);
            break;
        } else if (!block) {
            ret = 0;
			sceKernelSignalSema(q->mutex, 1);
            break;
        } else {
			sceKernelSignalSema(q->mutex, 1);
			sceKernelDelayThread(1000);
        }
    }
    return ret;
}

static inline void fill_rectangle(int x, int y, int w, int h, int color)
{
	my_rect area;
	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);

	area.x = x;
	area.y = y;
	area.w = w;
	area.h = h;
	my_clear_area(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, color);
}

static inline void fill_rectangle_alpha(int x, int y, int w, int h, int color, unsigned char alpha)
{
	my_rect area;
	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);

	area.x = x;
	area.y = y;
	area.w = w;
	area.h = h;
	color |= (alpha << 24);
	my_clear_area(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, color);
}

void MY_BlitSurface(unsigned int *dest, int dst_pitch, MY_Surface *surface, my_rect area)
{
	int w, h, i, pitch;
	int use_colorkey, colorkey;
	MY_Palette *palette;
	unsigned int *dst;

	w = (surface->w <= area.w) ? surface->w : area.w;
	h = (surface->h <= area.h) ? surface->h : area.h;
	
	if (surface->bpp == 8)
	{
		unsigned char *src;
		unsigned char r,g,b;

		palette = surface->palette;
		dst = &dest[area.x + area.y * dst_pitch];
		src = (unsigned char *)surface->pixels;
		pitch = surface->pitch;
		use_colorkey = surface->use_colorkey;
		colorkey = surface->colorkey;
		while (h)
		{
			for (i = 0; i < w; i++)
			{
				if (!use_colorkey || colorkey != src[i])
				{
					r = palette->colors[src[i]].r;
					g = palette->colors[src[i]].g;
					b = palette->colors[src[i]].b;

					dst[i] = RGB(r,g,b);
				}
			}
			dst += dst_pitch;
			src += pitch;
			h--;
		}
	}
	else
	{
		unsigned int *src;
		
		dst = &dest[area.x + area.y * dst_pitch];
		src = (unsigned int *)surface->pixels;
		pitch = surface->pitch / 4;

		while (h)
		{			
			for (i = 0; i < w; i++)
			{
				dst[i] = AlphaBlend(dst[i], src[i], (src[i] >> 24));
			}
			dst += dst_pitch;
			src += pitch;
			h--;
		}
	}
}

// 8bit only
void MY_BlitSurfaceAlpha(unsigned int *dest, int dst_pitch, MY_Surface *surface, my_rect area, unsigned char alpha)
{
	int w, h, i, pitch;
	int use_colorkey, colorkey;
	MY_Palette *palette;
	unsigned int *dst;

	w = (surface->w <= area.w) ? surface->w : area.w;
	h = (surface->h <= area.h) ? surface->h : area.h;
	
	if (surface->bpp == 8)
	{
		unsigned char *src;
		unsigned char r,g,b;

		palette = surface->palette;
		dst = &dest[area.x + area.y * dst_pitch];
		src = (unsigned char *)surface->pixels;
		pitch = surface->pitch;
		use_colorkey = surface->use_colorkey;
		colorkey = surface->colorkey;
		while (h)
		{
			for (i = 0; i < w; i++)
			{
				if (!use_colorkey || colorkey != src[i])
				{
					r = palette->colors[src[i]].r;
					g = palette->colors[src[i]].g;
					b = palette->colors[src[i]].b;

					dst[i] = RGBA(r,g,b,alpha);
				}
			}
			dst += dst_pitch;
			src += pitch;
			h--;
		}
	}
}

#ifdef SUBTITLE


#define SCALEBITS 10
#define ONE_HALF  (1 << (SCALEBITS - 1))
#define FIX(x)    ((int) ((x) * (1<<SCALEBITS) + 0.5))

#define RGB_TO_Y_CCIR(r, g, b) \
((FIX(0.29900*219.0/255.0) * (r) + FIX(0.58700*219.0/255.0) * (g) + \
  FIX(0.11400*219.0/255.0) * (b) + (ONE_HALF + (16 << SCALEBITS))) >> SCALEBITS)

#define RGB_TO_U_CCIR(r1, g1, b1, shift)\
(((- FIX(0.16874*224.0/255.0) * r1 - FIX(0.33126*224.0/255.0) * g1 +         \
     FIX(0.50000*224.0/255.0) * b1 + (ONE_HALF << shift) - 1) >> (SCALEBITS + shift)) + 128)

#define RGB_TO_V_CCIR(r1, g1, b1, shift)\
(((FIX(0.50000*224.0/255.0) * r1 - FIX(0.41869*224.0/255.0) * g1 -           \
   FIX(0.08131*224.0/255.0) * b1 + (ONE_HALF << shift) - 1) >> (SCALEBITS + shift)) + 128)

#define ALPHA_BLEND(a, oldp, newp, s)\
((((oldp << s) * (255 - (a))) + (newp * (a))) / (255 << s))

#define RGBA_IN(r, g, b, a, s)\
{\
    unsigned int v = ((const uint32_t *)(s))[0];\
    a = (v >> 24) & 0xff;\
    r = (v >> 16) & 0xff;\
    g = (v >> 8) & 0xff;\
    b = v & 0xff;\
}

#define YUVA_IN(y, u, v, a, s, pal)\
{\
    unsigned int val = ((const uint32_t *)(pal))[*(const uint8_t*)s];\
    a = (val >> 24) & 0xff;\
    y = (val >> 16) & 0xff;\
    u = (val >> 8) & 0xff;\
    v = val & 0xff;\
}

#define YUVA_OUT(d, y, u, v, a)\
{\
    ((uint32_t *)(d))[0] = (a << 24) | (y << 16) | (u << 8) | v;\
}


#define BPP 1

static void blend_subrect(AVPicture *dst, const AVSubtitleRect *rect)
{
    int wrap, wrap3, width2, skip2;
    int y, u, v, a, u1, v1, a1, w, h;
    uint8_t *lum, *cb, *cr;
    const uint8_t *p;
    const uint32_t *pal;

    lum = dst->data[0] + rect->y * dst->linesize[0];
    cb = dst->data[1] + (rect->y >> 1) * dst->linesize[1];
    cr = dst->data[2] + (rect->y >> 1) * dst->linesize[2];

    width2 = (rect->w + 1) >> 1;
    skip2 = rect->x >> 1;
    wrap = dst->linesize[0];
    wrap3 = rect->linesize;
    p = rect->bitmap;
    pal = rect->rgba_palette;  /* Now in YCrCb! */
    
    if (rect->y & 1) {
        lum += rect->x;
        cb += skip2;
        cr += skip2;
    
        if (rect->x & 1) {
            YUVA_IN(y, u, v, a, p, pal);
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);
            cb[0] = ALPHA_BLEND(a >> 2, cb[0], u, 0);
            cr[0] = ALPHA_BLEND(a >> 2, cr[0], v, 0);
            cb++;
            cr++;
            lum++;
            p += BPP;
        }
        for(w = rect->w - (rect->x & 1); w >= 2; w -= 2) {
            YUVA_IN(y, u, v, a, p, pal);
            u1 = u;
            v1 = v;
            a1 = a;
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);

            YUVA_IN(y, u, v, a, p + BPP, pal);
            u1 += u;
            v1 += v;
            a1 += a;
            lum[1] = ALPHA_BLEND(a, lum[1], y, 0);
            cb[0] = ALPHA_BLEND(a1 >> 2, cb[0], u1, 1);
            cr[0] = ALPHA_BLEND(a1 >> 2, cr[0], v1, 1);
            cb++;
            cr++;
            p += 2 * BPP;
            lum += 2;
        }
        if (w) {
            YUVA_IN(y, u, v, a, p, pal);
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);
            cb[0] = ALPHA_BLEND(a >> 2, cb[0], u, 0);
            cr[0] = ALPHA_BLEND(a >> 2, cr[0], v, 0);
        }
        p += wrap3 + (wrap3 - rect->w * BPP);
        lum += wrap + (wrap - rect->w - rect->x);
        cb += dst->linesize[1] - width2 - skip2;
        cr += dst->linesize[2] - width2 - skip2;
    }
    for(h = rect->h - (rect->y & 1); h >= 2; h -= 2) {
        lum += rect->x;
        cb += skip2;
        cr += skip2;
    
        if (rect->x & 1) {
            YUVA_IN(y, u, v, a, p, pal);
            u1 = u;
            v1 = v;
            a1 = a;
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);
            p += wrap3;
            lum += wrap;
            YUVA_IN(y, u, v, a, p, pal);
            u1 += u;
            v1 += v;
            a1 += a;
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);
            cb[0] = ALPHA_BLEND(a1 >> 2, cb[0], u1, 1);
            cr[0] = ALPHA_BLEND(a1 >> 2, cr[0], v1, 1);
            cb++;
            cr++;
            p += -wrap3 + BPP;
            lum += -wrap + 1;
        }
        for(w = rect->w - (rect->x & 1); w >= 2; w -= 2) {
            YUVA_IN(y, u, v, a, p, pal);
            u1 = u;
            v1 = v;
            a1 = a;
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);

            YUVA_IN(y, u, v, a, p, pal);
            u1 += u;
            v1 += v;
            a1 += a;
            lum[1] = ALPHA_BLEND(a, lum[1], y, 0);
            p += wrap3;
            lum += wrap;

            YUVA_IN(y, u, v, a, p, pal);
            u1 += u;
            v1 += v;
            a1 += a;
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);

            YUVA_IN(y, u, v, a, p, pal);
            u1 += u;
            v1 += v;
            a1 += a;
            lum[1] = ALPHA_BLEND(a, lum[1], y, 0);

            cb[0] = ALPHA_BLEND(a1 >> 2, cb[0], u1, 2);
            cr[0] = ALPHA_BLEND(a1 >> 2, cr[0], v1, 2);

            cb++;
            cr++;
            p += -wrap3 + 2 * BPP;
            lum += -wrap + 2;
        }
        if (w) {
            YUVA_IN(y, u, v, a, p, pal);
            u1 = u;
            v1 = v;
            a1 = a;
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);
            p += wrap3;
            lum += wrap;
            YUVA_IN(y, u, v, a, p, pal);
            u1 += u;
            v1 += v;
            a1 += a;
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);
            cb[0] = ALPHA_BLEND(a1 >> 2, cb[0], u1, 1);
            cr[0] = ALPHA_BLEND(a1 >> 2, cr[0], v1, 1);
            cb++;
            cr++;
            p += -wrap3 + BPP;
            lum += -wrap + 1;
        }
        p += wrap3 + (wrap3 - rect->w * BPP);
        lum += wrap + (wrap - rect->w - rect->x);
        cb += dst->linesize[1] - width2 - skip2;
        cr += dst->linesize[2] - width2 - skip2;
    }
    /* handle odd height */
    if (h) {
        lum += rect->x;
        cb += skip2;
        cr += skip2;
    
        if (rect->x & 1) {
            YUVA_IN(y, u, v, a, p, pal);
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);
            cb[0] = ALPHA_BLEND(a >> 2, cb[0], u, 0);
            cr[0] = ALPHA_BLEND(a >> 2, cr[0], v, 0);
            cb++;
            cr++;
            lum++;
            p += BPP;
        }
        for(w = rect->w - (rect->x & 1); w >= 2; w -= 2) {
            YUVA_IN(y, u, v, a, p, pal);
            u1 = u;
            v1 = v;
            a1 = a;
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);

            YUVA_IN(y, u, v, a, p + BPP, pal);
            u1 += u;
            v1 += v;
            a1 += a;
            lum[1] = ALPHA_BLEND(a, lum[1], y, 0);
            cb[0] = ALPHA_BLEND(a1 >> 2, cb[0], u, 1);
            cr[0] = ALPHA_BLEND(a1 >> 2, cr[0], v, 1);
            cb++;
            cr++;
            p += 2 * BPP;
            lum += 2;
        }
        if (w) {
            YUVA_IN(y, u, v, a, p, pal);
            lum[0] = ALPHA_BLEND(a, lum[0], y, 0);
            cb[0] = ALPHA_BLEND(a >> 2, cb[0], u, 0);
            cr[0] = ALPHA_BLEND(a >> 2, cr[0], v, 0);
        }
    }
}

static void free_subpicture(SubPicture *sp)
{
    int i;
    
    for (i = 0; i < sp->sub.num_rects; i++)
    {
        av_free(sp->sub.rects[i].bitmap);
        av_free(sp->sub.rects[i].rgba_palette);
    }
    
    av_free(sp->sub.rects);
    
    memset(&sp->sub, 0, sizeof(AVSubtitle));
}
#endif

static void video_image_display(VideoState *is)
{
    VideoPicture *vp;
	int src_w, src_h;
	int src_pitch;
	int dst_x, dst_y, dst_w, dst_h;
	u8 *tex;

    vp = &is->pictq[is->pictq_rindex];
    if (vp->bmp) {
//		if (display_clock)
//		{
//			my_rect area;
//			area.x = 0;
//			area.y = 0;
//			area.w = is->video_st->codec->width;
//			area.h = is->video_st->codec->height;
//
//			if (!use_ttf_font)
//			{
//				my_draw_text((u32*)vp->bmp->data[0], 512, 0, 0, (const u8 *)clock_text, 11, MY_COLOR_RED, area);
//			}
//			else
//			{
//				MY_Surface* textbuf;
//				MY_Color fg = {0xFF, 0x00, 0x00, 0};
//
//   				ascii2unicode((u8 *)clock_text, unicode_text);
//				textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, fg);
//				MY_BlitSurface((u32*)vp->bmp->data[0], 512, textbuf, area);
//				MY_FreeSurface(textbuf);
//			}
//		}
#ifdef JONNY_YUV2RGB
		guDrawBuffer((u32*)vp->bmp->data[0], is->video_st->codec->width, is->video_st->codec->height, 512, 0, 
					disp_w, disp_h);
#else
		tex = vp->bmp->planes[0];
		src_w = is->video_st->codec->width;
		src_h = is->video_st->codec->height;
		src_pitch = vp->bmp->pitches[0];

		dst_w = disp_w;
		dst_h = disp_h;

		dst_x = (480 - dst_w) / 2;
		dst_y = (272 - dst_h) / 2;

		pgBeginRender();
		blit_y_plane(tex, src_w, src_h, src_pitch, dst_x, dst_y, dst_w, dst_h);

		tex = vp->bmp->planes[2];
		src_w /= 2;
		src_h /= 2;
		src_pitch = vp->bmp->pitches[2];
		blit_uv_plane(tex, src_w, src_h, src_pitch, dst_x, dst_y, dst_w, dst_h, 0, add_u_clut);
		blit_uv_plane(tex, src_w, src_h, src_pitch, dst_x, dst_y, dst_w, dst_h, 1, sub_u_clut);

		tex = vp->bmp->planes[1];
		src_pitch = vp->bmp->pitches[1];
		blit_uv_plane(tex, src_w, src_h, src_pitch, dst_x, dst_y, dst_w, dst_h, 0, add_v_clut);
		blit_uv_plane(tex, src_w, src_h, src_pitch, dst_x, dst_y, dst_w, dst_h, 1, sub_v_clut);

		pgEndRender();
#endif
		disp_update();
		scePowerTick(0);
    } else {
    }
}

static inline int compute_mod(int a, int b)
{
    a = a % b;
    if (a >= 0) 
        return a;
    else
        return a + b;
}

static void video_audio_display(VideoState *s)
{
#if 1
    int i, i_start, x, y1, y, ys, delay, n, nb_display_channels;
    int ch, channels, h, h2, bgcolor, fgcolor;
    int16_t time_diff;

	char *txt;
	int len;
	my_rect area;

	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);

	fill_rectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, 0x00000000);

//	sceKernelWaitSema(s->audio_buffer_mutex, 1, 0);
    /* compute display index : center on currently output samples */
    channels = s->audio_st->codec->channels;
    nb_display_channels = channels;
    if (!s->paused) {
        n = 2 * channels;
        delay = audio_write_get_buf_size(s);
        delay /= n;
        
        /* to be more precise, we take into account the time spent since
           the last buffer computation */
        if (audio_callback_time) {
            time_diff = av_gettime() - audio_callback_time;
//            delay += (time_diff * s->audio_st->codec->sample_rate) / 1000000;
            delay += (time_diff * 44100) / 1000000; // Jini ... experimenting audio sync 051117
        }
        
        delay -= s->width / 2;
        if (delay < s->width)
            delay = s->width;
        i_start = compute_mod(s->sample_array_index - delay * channels, SAMPLE_ARRAY_SIZE);
        s->last_i_start = i_start;
    } else {
        i_start = s->last_i_start;
    }

	bgcolor = MY_COLOR_BLACK;
    fill_rectangle(s->xleft, s->ytop, s->width, s->height, bgcolor);
	fgcolor = MY_COLOR_WHITE;

    /* total height for one channel */
    h = s->height / nb_display_channels;
    /* graph height / 2 */
    h2 = (h * 9) / 20;
    for(ch = 0;ch < nb_display_channels; ch++) {
        i = i_start + ch;
        y1 = s->ytop + ch * h + (h / 2); /* position of center line */
        for(x = 0; x < s->width; x++) {
            y = (s->sample_array[i] * h2) >> 15;
            if (y < 0) {
                y = -y;
                ys = y1 - y;
            } else {
                ys = y1;
            }
            fill_rectangle(s->xleft + x, ys, 1, y, fgcolor);
            i += channels;
            if (i >= SAMPLE_ARRAY_SIZE)
                i -= SAMPLE_ARRAY_SIZE;
        }
    }
//	sceKernelSignalSema(s->audio_buffer_mutex, 1);

	area.x = 0;
	area.y = 0;
	area.w = SCREEN_WIDTH;
	area.h = SCREEN_HEIGHT;

	y = SCREEN_HEIGHT / 2 - 20;

	txt = s->ic->title;
	len = strlen((const char *)txt);
	if (len)
	{
		if (!use_ttf_font)
		{
			my_draw_aligned_text(screen_buffer, LINESIZE, y, (const u8 *)txt, len, ALIGN_MIDDLE, MY_COLOR_RED, area);
		}
		else
		{
			MY_Surface* textbuf;
			MY_Color fg = {0xFF, 0x00, 0x00, 0};
			my_rect msg_area;

		   	ascii2unicode((u8 *)txt, unicode_text);
			textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, fg);
			msg_area.x = area.x + (area.w - textbuf->w) / 2;
			msg_area.y = area.y + y;
			msg_area.w = textbuf->w;
			msg_area.h = textbuf->h;
			MY_BlitSurface(screen_buffer, LINESIZE, textbuf, msg_area);
			MY_FreeSurface(textbuf);
		}
		y += 20;
	}
	else
	{
		txt = media_file_list[media_idx].d_name;
		len = strlen((const char *)txt);
		if (!use_ttf_font)
		{
			my_draw_aligned_text(screen_buffer, LINESIZE, y, (const u8 *)txt, len, ALIGN_MIDDLE, MY_COLOR_RED, area);
		}
		else
		{
			MY_Surface* textbuf;
			MY_Color fg = {0xFF, 0x00, 0x00, 0};
			my_rect msg_area;

		   	ascii2unicode((u8 *)txt, unicode_text);
			textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, fg);
			msg_area.x = area.x + (area.w - textbuf->w) / 2;
			msg_area.y = area.y + y;
			msg_area.w = textbuf->w;
			msg_area.h = textbuf->h;
			MY_BlitSurface(screen_buffer, LINESIZE, textbuf, msg_area);
			MY_FreeSurface(textbuf);
		}
		y += 20;
	}
	txt = s->ic->author;
	len = strlen((const char *)txt);
	if (len)
	{
		if (!use_ttf_font)
		{
			my_draw_aligned_text(screen_buffer, LINESIZE, y, (const u8 *)txt, len, ALIGN_MIDDLE, MY_COLOR_ORANGE, area);
		}
		else
		{
			MY_Surface* textbuf;
			MY_Color fg = {0xFF, 0x5A, 0x00, 0};
			my_rect msg_area;

		   	ascii2unicode((u8 *)txt, unicode_text);
			textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, fg);
			msg_area.x = area.x + (area.w - textbuf->w) / 2;
			msg_area.y = area.y + y;
			msg_area.w = textbuf->w;
			msg_area.h = textbuf->h;
			MY_BlitSurface(screen_buffer, LINESIZE, textbuf, msg_area);
			MY_FreeSurface(textbuf);
		}
		y += 20;
	}
	disp_copy((char *)screen_buffer, LINESIZE, 0, 0, LINESIZE, SCREEN_HEIGHT, 0, 0);
	disp_update();	
#endif
}

/* display the current picture, if any */
static void video_display(VideoState *is)
{
    if (is->audio_st && is->show_audio) 
        video_audio_display(is);
    else if (is->video_st)
        video_image_display(is);
}

#define MY_EVENT_REFRESH 0x1000000

static void refresh_timer_cb(void *priv)
{
	add_event(MY_EVENT_REFRESH, priv);
}

/* schedule a video refresh in 'delay' ms */
static void schedule_refresh(VideoState *is, int delay)
{
//	next_refresh_time = delay * 1000 + av_gettime();
//	need_timer_check = 1;
	refresh_timer_id = set_timer(delay, refresh_timer_cb, is, 0);
}

/* get the current audio clock value */
static double get_audio_clock(VideoState *is)
{
    double pts;
    int hw_buf_size, bytes_per_sec;
    pts = is->audio_clock;
    hw_buf_size = audio_write_get_buf_size(is);
    bytes_per_sec = 0;
    if (is->audio_st) {
//        bytes_per_sec = is->audio_st->codec->sample_rate * 
//            2 * is->audio_st->codec->channels;
        bytes_per_sec = 44100 * 
            2 * is->audio_st->codec->channels; // Jini ... experimenting audio sync 051117
    }
    if (bytes_per_sec)
        pts -= (double)hw_buf_size / bytes_per_sec;
    return pts;
}

/* get the current video clock value */
static double get_video_clock(VideoState *is)
{
    double delta;
    if (is->paused) {
        delta = 0;
    } else {
        delta = (av_gettime() - is->video_current_pts_time) / 1000000.0;
    }
    return is->video_current_pts + delta;
}

/* get the current external clock value */
static double get_external_clock(VideoState *is)
{
    int64_t ti;
    ti = av_gettime();
    return is->external_clock + ((ti - is->external_clock_time) * 1e-6);
}

/* get the current master clock value */
static double get_master_clock(VideoState *is)
{
    double val;

    if (is->av_sync_type == AV_SYNC_VIDEO_MASTER) {
        if (is->video_st)
            val = get_video_clock(is);
        else
            val = get_audio_clock(is);
    } else if (is->av_sync_type == AV_SYNC_AUDIO_MASTER) {
        if (is->audio_st)
            val = get_audio_clock(is);
        else
            val = get_video_clock(is);
    } else {
        val = get_external_clock(is);
    }
    return val;
}

/* seek in the stream */
static void stream_seek(VideoState *is, int64_t pos, int rel)
{
   if (!is->seek_req) {
        is->seek_pos = pos;
        is->seek_flags = rel < 0 ? AVSEEK_FLAG_BACKWARD : 0;
        is->seek_req = 1;
    }
}

/* pause or resume the video */
static void stream_pause(VideoState *is)
{
    is->paused = !is->paused;
    if (is->paused) {
		audio_pause();
        is->video_current_pts = get_video_clock(is);
    }
	else
	{
		audio_pause();
	}
}

/* called to display each frame */
static void video_refresh_timer(void *opaque)
{
    VideoState *is = opaque;
    VideoPicture *vp;
    double actual_delay, delay, sync_threshold, ref_clock, diff;


    if (is->video_st) {
        if (is->pictq_size == 0) {
            /* if no picture, need to wait */
            schedule_refresh(is, 1);
        } else {
//            video_display(is);
            /* dequeue the picture */
            vp = &is->pictq[is->pictq_rindex];

            /* update current video pts */
            is->video_current_pts = vp->pts;
            is->video_current_pts_time = av_gettime();

            /* compute nominal delay */
            delay = vp->pts - is->frame_last_pts;
            if (delay <= 0 || delay >= 1.0) {
                /* if incorrect delay, use previous one */
                delay = is->frame_last_delay;
            }
            is->frame_last_delay = delay;
            is->frame_last_pts = vp->pts;

            /* update delay to follow master synchronisation source */
            if (((is->av_sync_type == AV_SYNC_AUDIO_MASTER && is->audio_st) ||
                 is->av_sync_type == AV_SYNC_EXTERNAL_CLOCK)) {
                /* if video is slave, we try to correct big delays by
                   duplicating or deleting a frame */
                ref_clock = get_master_clock(is);
                diff = vp->pts - ref_clock;
                
                /* skip or repeat frame. We take into account the
                   delay to compute the threshold. I still don't know
                   if it is the best guess */
                sync_threshold = AV_SYNC_THRESHOLD;
                if (delay > sync_threshold)
                    sync_threshold = delay;
                if (fabs(diff) < AV_NOSYNC_THRESHOLD) {
                    if (diff <= -sync_threshold)
                        delay = 0;
                    else if (diff >= sync_threshold)
                        delay = 2 * delay;
                }
            }

            is->frame_timer += delay;
            /* compute the REAL delay (we need to do that to avoid
               long term errors */
            actual_delay = is->frame_timer - (av_gettime() / 1000000.0);
            if (actual_delay < 0.010) {
                /* XXX: should skip picture */
				skip_frames = 1;
                actual_delay = 0.010;
            }
            /* launch timer for next picture */
            schedule_refresh(is, (int)(actual_delay * 1000 + 0.5));

#if defined(DEBUG_SYNC)
            printf("video: delay=%0.3f actual_delay=%0.3f pts=%0.3f A-V=%f\n", 
                   delay, actual_delay, vp->pts, -diff);
#endif

            /* display picture */
            video_display(is);

            /* update queue size and signal for next picture */
            if (++is->pictq_rindex == VIDEO_PICTURE_QUEUE_SIZE)
                is->pictq_rindex = 0;
            
			sceKernelWaitSema(is->pictq_mutex, 1, 0);
            is->pictq_size--;
			sceKernelSignalSema(is->pictq_mutex, 1);
        }
    } else if (is->audio_st) {
        /* draw the next audio frame */

        schedule_refresh(is, 40);

        /* if only audio stream, then display the audio bars (better
           than nothing, just to test the implementation */
        
        /* display picture */
        video_display(is);
    } else {
        schedule_refresh(is, 100);
    }
    if (show_status) {
        static int64_t last_time;
        int64_t cur_time;
        int aqsize, vqsize;
        double av_diff;
        
        cur_time = av_gettime();
        if (!last_time || (cur_time - last_time) >= 500 * 1000) {
            aqsize = 0;
            vqsize = 0;
            if (is->audio_st)
                aqsize = is->audioq.size;
            if (is->video_st)
                vqsize = is->videoq.size;
            av_diff = 0;
            if (is->audio_st && is->video_st)
                av_diff = get_audio_clock(is) - get_video_clock(is);
            printf("%7.2f A-V:%7.3f aq=%5dKB vq=%5dKB    \n", 
                   get_master_clock(is), av_diff, aqsize / 1024, vqsize / 1024);
            fflush(stdout);
            last_time = cur_time;
        }
    }
}

/* allocate a picture (needs to do that in main thread to avoid
   potential locking problems */
static void alloc_picture(void *opaque)
{
    VideoState *is = opaque;
    VideoPicture *vp;

    vp = &is->pictq[is->pictq_windex];

    if (vp->bmp)
	{
#ifdef JONNY_YUV2RGB
		if (vp->buffer)
		{
			av_free(vp->buffer);
		}
		av_free(vp->bmp);
#else
		free_overlay_surface(vp->bmp);
#endif
	}

    // Allocate an AVFrame structure
#ifdef JONNY_YUV2RGB
    vp->bmp = avcodec_alloc_frame();
    if(vp->bmp)
	{
		int numBytes;
	    numBytes=avpicture_get_size(PIX_FMT_RGBA32, 512, is->video_st->codec->height);
		vp->buffer = malloc_64(4 * 512 * is->video_st->codec->height); // Jini 
		avpicture_fill((AVPicture *)vp->bmp, vp->buffer, PIX_FMT_RGBA32, 512, is->video_st->codec->height);
	}
#else
	vp->bmp = create_overlay_surface(is->video_st->codec->width, is->video_st->codec->height);
#endif	    	
    vp->width = is->video_st->codec->width;
    vp->height = is->video_st->codec->height;

	sceKernelWaitSema(is->pictq_mutex, 1, 0);
    vp->allocated = 1;
	sceKernelSignalSema(is->pictq_mutex, 1);
}

static int queue_picture(VideoState *is, AVFrame *src_frame, double pts)
{
    VideoPicture *vp;
    int dst_pix_fmt;
    AVPicture pict;
    
    /* wait until we have space to put a new picture */
	while (1)
	{
		sceKernelWaitSema(is->pictq_mutex, 1, 0);
		if (is->pictq_size >= VIDEO_PICTURE_QUEUE_SIZE && !is->videoq.abort_request)
		{
			sceKernelSignalSema(is->pictq_mutex, 1);
			sceKernelDelayThread(1000);
		}
		else
		{
			sceKernelSignalSema(is->pictq_mutex, 1);
			break;
		}
	}
    
    if (is->videoq.abort_request)
        return -1;

    vp = &is->pictq[is->pictq_windex];

    /* alloc or resize hardware picture buffer */
    if (!vp->bmp || 
        vp->width != is->video_st->codec->width ||
        vp->height != is->video_st->codec->height) {

        vp->allocated = 0;

        /* the allocation must be done in the main thread to avoid
           locking problems */
		need_alloc = 1;
        
        /* wait until the picture is allocated */
		while (1)
		{
			sceKernelWaitSema(is->pictq_mutex, 1, 0);
			if (!vp->allocated && !is->videoq.abort_request)
			{
				sceKernelSignalSema(is->pictq_mutex, 1);
				sceKernelDelayThread(1000);
			}
			else
			{
				sceKernelSignalSema(is->pictq_mutex, 1);
				break;
			}
		}

        if (is->videoq.abort_request)
            return -1;
    }

    /* if the frame is not skipped, then display it */
    if (vp->bmp) {
        /* get a pointer on the bitmap */
#ifdef JONNY_YUV2RGB
        me_csc                           // [jonny]
			(                            // [jonny]
			src_frame->data[0],          // [jonny]
			src_frame->data[1],          // [jonny]
			src_frame->data[2],          // [jonny]
			src_frame->linesize[0],      // [jonny]
			src_frame->linesize[1],      // [jonny]
			src_frame->linesize[2],      // [jonny]
			vp->buffer,                  // [jonny]
			is->video_st->codec->width,  // [jonny]
			is->video_st->codec->height, // [jonny]
			512                          // [jonny]
			);                           // [jonny]
#else
        dst_pix_fmt = PIX_FMT_YUV420P;
        pict.data[0] = vp->bmp->planes[0];
        pict.data[1] = vp->bmp->planes[2];
        pict.data[2] = vp->bmp->planes[1];

        pict.linesize[0] = vp->bmp->pitches[0];
        pict.linesize[1] = vp->bmp->pitches[2];
        pict.linesize[2] = vp->bmp->pitches[1];
        img_convert(&pict, dst_pix_fmt, 
                    (AVPicture *)src_frame, is->video_st->codec->pix_fmt, 
                    is->video_st->codec->width, is->video_st->codec->height);
#endif
        vp->pts = pts;

        /* now we can update the picture count */
        if (++is->pictq_windex == VIDEO_PICTURE_QUEUE_SIZE)
            is->pictq_windex = 0;
		sceKernelWaitSema(is->pictq_mutex, 1, 0);
        is->pictq_size++;
		sceKernelSignalSema(is->pictq_mutex, 1);
    }
    return 0;
}

/* compute the exact PTS for the picture if it is omitted in the stream */
static int output_picture2(VideoState *is, AVFrame *src_frame, double pts1)
{
    double frame_delay, pts;
    
    pts = pts1;

    if (pts != 0) {
        /* update video clock with pts, if present */
        is->video_clock = pts;
    } else {
        pts = is->video_clock;
    }
    /* update video clock for next frame */
	frame_delay = av_q2d(is->video_st->codec->time_base);

    /* for MPEG2, the frame can be repeated, so we update the
       clock accordingly */
    frame_delay += src_frame->repeat_pict * (frame_delay * 0.5);
    is->video_clock += frame_delay;

#if defined(DEBUG_SYNC) && 0
    {
        int ftype;
        if (src_frame->pict_type == FF_B_TYPE)
            ftype = 'B';
        else if (src_frame->pict_type == FF_I_TYPE)
            ftype = 'I';
        else
            ftype = 'P';
        printf("frame_type=%c clock=%0.3f pts=%0.3f\n", 
               ftype, pts, pts1);
    }
#endif
	if (!skip_frames)
	{
    	return queue_picture(is, src_frame, pts);
	}
	else
	{
		skip_frames--;
		return 0;
	}
}

static int video_thread(SceSize args, void *argp)
{
    VideoState *is = *(VideoState**)argp;
    AVPacket pkt1, *pkt = &pkt1;
    int len1, got_picture;
    AVFrame *frame= avcodec_alloc_frame();
    double pts;

    for(;;) {
        while (is->paused && !is->videoq.abort_request) {
			sceKernelDelayThread(10000);
        }
        if (packet_queue_get(&is->videoq, pkt, 1) < 0)
            break;
        /* NOTE: ipts is the PTS of the _first_ picture beginning in
           this packet, if any */
        pts = 0;
        if (pkt->dts != AV_NOPTS_VALUE)
            pts = av_q2d(is->video_st->time_base)*pkt->dts;

			sceKernelWaitSema(is->video_decoder_mutex, 1, 0);
            len1 = avcodec_decode_video(is->video_st->codec, 
                                        frame, &got_picture, 
                                        pkt->data, pkt->size);
			sceKernelSignalSema(is->video_decoder_mutex, 1);
            if (got_picture) {
//				if (!skip_frames)
//				{
                	if (output_picture2(is, frame, pts) < 0)
                    	goto the_end;
//				}
//				else
//				{
//					skip_frames--;
//				}
            }
        av_free_packet(pkt);
        if (step) 
            if (cur_stream)
                stream_pause(cur_stream);
    }
the_end:
    av_free(frame);
    return 0;
}

#ifdef SUBTITLE
static int subtitle_thread(void *arg)
{
    VideoState *is = arg;
    SubPicture *sp;
    AVPacket pkt1, *pkt = &pkt1;
    int len1, got_subtitle;
    double pts;
    int i, j;
    int r, g, b, y, u, v, a;

    for(;;) {
        while (is->paused && !is->subtitleq.abort_request) {
            SDL_Delay(10);
        }
        if (packet_queue_get(&is->subtitleq, pkt, 1) < 0)
            break;
            
        SDL_LockMutex(is->subpq_mutex);
        while (is->subpq_size >= SUBPICTURE_QUEUE_SIZE &&
               !is->subtitleq.abort_request) {
            SDL_CondWait(is->subpq_cond, is->subpq_mutex);
        }
        SDL_UnlockMutex(is->subpq_mutex);
        
        if (is->subtitleq.abort_request)
            goto the_end;
            
        sp = &is->subpq[is->subpq_windex];

       /* NOTE: ipts is the PTS of the _first_ picture beginning in
           this packet, if any */
        pts = 0;
        if (pkt->pts != AV_NOPTS_VALUE)
            pts = av_q2d(is->subtitle_st->time_base)*pkt->pts;

        SDL_LockMutex(is->subtitle_decoder_mutex);
        len1 = avcodec_decode_subtitle(is->subtitle_st->codec, 
                                    &sp->sub, &got_subtitle, 
                                    pkt->data, pkt->size);
        SDL_UnlockMutex(is->subtitle_decoder_mutex);
//            if (len1 < 0)
//                break;
        if (got_subtitle && sp->sub.format == 0) {
            sp->pts = pts;
            
            for (i = 0; i < sp->sub.num_rects; i++)
            {
                for (j = 0; j < sp->sub.rects[i].nb_colors; j++)
                {
                    RGBA_IN(r, g, b, a, sp->sub.rects[i].rgba_palette + j);
                    y = RGB_TO_Y_CCIR(r, g, b);
                    u = RGB_TO_U_CCIR(r, g, b, 0);
                    v = RGB_TO_V_CCIR(r, g, b, 0);
                    YUVA_OUT(sp->sub.rects[i].rgba_palette + j, y, u, v, a);
                }
            }

            /* now we can update the picture count */
            if (++is->subpq_windex == SUBPICTURE_QUEUE_SIZE)
                is->subpq_windex = 0;
            SDL_LockMutex(is->subpq_mutex);
            is->subpq_size++;
            SDL_UnlockMutex(is->subpq_mutex);
        }
        av_free_packet(pkt);
//        if (step) 
//            if (cur_stream)
//                stream_pause(cur_stream);
    }
 the_end:
    return 0;
}
#endif

/* copy samples for viewing in editor window */
static void update_sample_display(VideoState *is, short *samples, int samples_size)
{
    int size, len, channels;

    channels = is->audio_st->codec->channels;

    size = samples_size / sizeof(short);
    while (size > 0) {
        len = SAMPLE_ARRAY_SIZE - is->sample_array_index;
        if (len > size)
            len = size;
        memcpy(is->sample_array + is->sample_array_index, samples, len * sizeof(short));
        samples += len;
        is->sample_array_index += len;
        if (is->sample_array_index >= SAMPLE_ARRAY_SIZE)
            is->sample_array_index = 0;
        size -= len;
    }
}

/* return the new audio buffer size (samples can be added or deleted
   to get better sync if video or external master clock) */
static int synchronize_audio(VideoState *is, short *samples, 
                             int samples_size1, double pts)
{
    int n, samples_size;
    double ref_clock;
    
    n = 2 * is->audio_st->codec->channels;
    samples_size = samples_size1;

    /* if not master, then we try to remove or add samples to correct the clock */
    if (((is->av_sync_type == AV_SYNC_VIDEO_MASTER && is->video_st) ||
         is->av_sync_type == AV_SYNC_EXTERNAL_CLOCK)) {
        double diff, avg_diff;
        int wanted_size, min_size, max_size, nb_samples;
            
        ref_clock = get_master_clock(is);
        diff = get_audio_clock(is) - ref_clock;
        
        if (diff < AV_NOSYNC_THRESHOLD) {
            is->audio_diff_cum = diff + is->audio_diff_avg_coef * is->audio_diff_cum;
            if (is->audio_diff_avg_count < AUDIO_DIFF_AVG_NB) {
                /* not enough measures to have a correct estimate */
                is->audio_diff_avg_count++;
            } else {
                /* estimate the A-V difference */
                avg_diff = is->audio_diff_cum * (1.0 - is->audio_diff_avg_coef);

                if (fabs(avg_diff) >= is->audio_diff_threshold) {
//                    wanted_size = samples_size + ((int)(diff * is->audio_st->codec->sample_rate) * n);
                    wanted_size = samples_size + ((int)(diff * 44100) * n); // Jini ... experimenting audio sync 051117
                    nb_samples = samples_size / n;
                
                    min_size = ((nb_samples * (100 - SAMPLE_CORRECTION_PERCENT_MAX)) / 100) * n;
                    max_size = ((nb_samples * (100 + SAMPLE_CORRECTION_PERCENT_MAX)) / 100) * n;
                    if (wanted_size < min_size)
                        wanted_size = min_size;
                    else if (wanted_size > max_size)
                        wanted_size = max_size;
                    
                    /* add or remove samples to correction the synchro */
                    if (wanted_size < samples_size) {
                        /* remove samples */
                        samples_size = wanted_size;
                    } else if (wanted_size > samples_size) {
                        uint8_t *samples_end, *q;
                        int nb;
                        
                        /* add samples */
                        nb = (samples_size - wanted_size);
                        samples_end = (uint8_t *)samples + samples_size - n;
                        q = samples_end + n;
                        while (nb > 0) {
                            memcpy(q, samples_end, n);
                            q += n;
                            nb -= n;
                        }
                        samples_size = wanted_size;
                    }
                }
            }
        } else {
            /* too big difference : may be initial PTS errors, so
               reset A-V filter */
            is->audio_diff_avg_count = 0;
            is->audio_diff_cum = 0;
        }
    }

    return samples_size;
}

/* decode one audio frame and returns its uncompressed size */
static int audio_decode_frame(VideoState *is, uint8_t *audio_buf, double *pts_ptr)
{
    AVPacket *pkt = &is->audio_pkt;
    int n, len1, data_size;
    double pts;

    for(;;) {
        /* NOTE: the audio packet can contain several frames */
        while (is->audio_pkt_size > 0) {
			sceKernelWaitSema(is->audio_decoder_mutex, 1, 0);
            len1 = avcodec_decode_audio(is->audio_st->codec, 
                                        (int16_t *)audio_buf, &data_size, 
                                        is->audio_pkt_data, is->audio_pkt_size);
			sceKernelSignalSema(is->audio_decoder_mutex, 1);

            if (len1 < 0) {
                /* if error, we skip the frame */
                is->audio_pkt_size = 0;
                break;
            }
            
            is->audio_pkt_data += len1;
            is->audio_pkt_size -= len1;
            if (data_size <= 0)
                continue;
            /* if no pts, then compute it */
            pts = is->audio_clock;
            *pts_ptr = pts;

//            n = 2 * is->audio_st->codec->channels;
//            is->audio_clock += (double)data_size / 
//                (double)(n * is->audio_st->codec->sample_rate);

#if defined(DEBUG_SYNC)
            {
                static double last_clock;
                printf("audio: delay=%0.3f clock=%0.3f pts=%0.3f\n",
                       is->audio_clock - last_clock,
                       is->audio_clock, pts);
                last_clock = is->audio_clock;
            }
#endif
            return data_size;
        }

        /* free the current packet */
        if (pkt->data)
		{
            av_free_packet(pkt);
        }
        if (is->paused || is->audioq.abort_request) {
            return -1;
        }
        
        /* read next packet */
        if (packet_queue_get(&is->audioq, pkt, 1) < 0)
            return -1;
        is->audio_pkt_data = pkt->data;
        is->audio_pkt_size = pkt->size;
        
        /* if update the audio clock with the pts */
        if (pkt->pts != AV_NOPTS_VALUE) {
            is->audio_clock = av_q2d(is->audio_st->time_base)*pkt->pts;
        }
    }
}

/* get the current audio output buffer size, in samples. With SDL, we
   cannot have a precise information */
static int audio_write_get_buf_size(VideoState *is)
{
    return is->audio_hw_buf_size - is->audio_buf_index;
}


/* prepare a new audio buffer */
#define MAX_AUDIO_PACKET_SIZE (128 * 1024)
//static uint16_t audio_buf[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2];
static uint16_t audio_buf[AVCODEC_MAX_AUDIO_FRAME_SIZE * 2];

void sdl_audio_callback(short* buffer, unsigned long length)
{
	uint8_t *stream;
	int len;
	int num_ch;
	int i, j;
    VideoState *is = cur_stream;
    int audio_size, len1;
    double pts;
	int size_out;
	int n; // Jini ... experimenting audio sync 051117

	stream = (uint8_t *)buffer;
	if (!is->audio_st)
	{
		memset(stream, 0, length);
		return;
	}
	num_ch = is->audio_st->codec->channels;

	len = length;

    audio_callback_time = av_gettime();
   sceKernelWaitSema(is->audio_buffer_mutex, 1, 0);
    while (len > 0) {
        if (is->audio_buf_index >= is->audio_buf_size) {

		   size_out = audio_decode_frame(is, audio_buf, &pts);
		   if (size_out > 0)
		   {
		     if (is->audio_resample)
		     {
				audio_size = audio_resample(is->resample, (short *)is->audio_buf, (short *)audio_buf, 
										  size_out / (is->audio_st->codec->channels * 2));
				audio_size = audio_size * 2 * 2;
		     }
		     else
		     {
				audio_size = size_out;
				memcpy(is->audio_buf, audio_buf, size_out);
		     }
		   }
		   else
		   {
			 audio_size = size_out;
		   }
// Jini... experimenting audio sync 051117
		   n = 2 * is->audio_st->codec->channels;
           is->audio_clock += (double)audio_size / (double)(n * 44100);

           if (audio_size < 0) {
                /* if error, just output silence */
               is->audio_buf_size = 1024;
               memset(is->audio_buf, 0, is->audio_buf_size);
           } else {
               if (is->show_audio)
                   update_sample_display(is, (int16_t *)is->audio_buf, audio_size);
               audio_size = synchronize_audio(is, (int16_t *)is->audio_buf, audio_size, 
                                              pts);
               is->audio_buf_size = audio_size;
           }
           is->audio_buf_index = 0;
        }

        len1 = is->audio_buf_size - is->audio_buf_index;
        if (len1 > len)
            len1 = len;

       	memcpy(stream, (uint8_t *)is->audio_buf + is->audio_buf_index, len1);
        len -= len1;
       	stream += len1;
        is->audio_buf_index += len1;
    }
   	sceKernelSignalSema(is->audio_buffer_mutex, 1);
}

/* open a given stream. Return 0 if OK */
static int stream_component_open(VideoState *is, int stream_index)
{
    AVFormatContext *ic = is->ic;
    AVCodecContext *enc;
    AVCodec *codec;

    if (stream_index < 0 || stream_index >= ic->nb_streams)
        return -1;
    enc = ic->streams[stream_index]->codec;
    
    /* prepare audio output */
    if (enc->codec_type == CODEC_TYPE_AUDIO) {
//		if (enc->sample_rate < 44100)
		if (enc->sample_rate != 44100) // Jini .. 051117
		{
			const char *txt;
			txt = (const char *)mytxt_get_string(MY_TXT_INIT_SOUND, language);
//			show_info_box(txt);

			is->audio_resample = 1;
     		if(is->audio_resample){
            	is->resample = audio_resample_init(2, enc->channels,
	                                                44100, enc->sample_rate);
			}
		}
		else
		{
			is->audio_resample = 0;
		}

//		pspAudioSetChannelCallback(0, sdl_audio_callback, NULL);
		init_audio(enc->channels, sdl_audio_callback);
        is->audio_hw_buf_size = 1024 * 2 * 2;
    }																	   

    codec = avcodec_find_decoder(enc->codec_id);
    enc->debug_mv = debug_mv;
    enc->debug = debug;
    if(debug)
        av_log_set_level(AV_LOG_DEBUG);
    enc->workaround_bugs = workaround_bugs;
	enc->lowres = lowres;
    if(lowres) enc->flags |= CODEC_FLAG_EMU_EDGE;
    enc->idct_algo= idct;
    if(fast) enc->flags2 |= CODEC_FLAG2_FAST;
    enc->skip_frame= skip_frame;
    enc->skip_idct= skip_idct;
    enc->skip_loop_filter= skip_loop_filter;
    enc->error_resilience= error_resilience;
    enc->error_concealment= error_concealment;
    if (!codec ||
        avcodec_open(enc, codec) < 0)
        return -1;
#if defined(HAVE_PTHREADS) || defined(HAVE_W32THREADS) || defined(HAVE_PSPTHREADS)
    if(thread_count>1)
        avcodec_thread_init(enc, thread_count);
#endif
    enc->thread_count= thread_count;
    switch(enc->codec_type) {
    case CODEC_TYPE_AUDIO:
        is->audio_stream = stream_index;
        is->audio_st = ic->streams[stream_index];
        is->audio_buf_size = 0;
        is->audio_buf_index = 0;

        /* init averaging filter */
        is->audio_diff_avg_coef = exp(log(0.01) / AUDIO_DIFF_AVG_NB);
        is->audio_diff_avg_count = 0;
        /* since we do not have a precise anough audio fifo fullness,
           we correct audio sync only if larger than this threshold */
        is->audio_diff_threshold = 2.0 * SDL_AUDIO_BUFFER_SIZE / enc->sample_rate;

        memset(&is->audio_pkt, 0, sizeof(is->audio_pkt));
        packet_queue_init(&is->audioq);
        break;
    case CODEC_TYPE_VIDEO:
		if (ic->streams[stream_index]->codec->width > 480)
		{
			return -1;
		}

		scePowerSetClockFrequency(333,333,166);
        is->video_stream = stream_index;
        is->video_st = ic->streams[stream_index];

        is->frame_last_delay = 40e-3;
        is->frame_timer = (double)av_gettime() / 1000000.0;
        is->video_current_pts_time = av_gettime();

        packet_queue_init(&is->videoq);
		set_disp_mode(is, mode);
		skip_frames = 0;
//        is->video_tid = sceKernelCreateThread("Video thread", (SceKernelThreadEntry)video_thread, 0x13, 0x20000, THREAD_ATTR_VFPU, 0);
        is->video_tid = sceKernelCreateThread("Video thread", (SceKernelThreadEntry)video_thread, 0x14, 0x20000, THREAD_ATTR_VFPU, 0);
		if (is->video_tid >= 0)
		{
			sceKernelStartThread(is->video_tid, sizeof(VideoState *), &is);
		}
        break;
#ifdef SUBTITLE
    case CODEC_TYPE_SUBTITLE:
        is->subtitle_stream = stream_index;
        is->subtitle_st = ic->streams[stream_index];
        packet_queue_init(&is->subtitleq);
        
        is->subtitle_tid = SDL_CreateThread(subtitle_thread, is);
        break;
#endif
    default:
        break;
    }
    return 0;
}

static void stream_component_close(VideoState *is, int stream_index)
{
    AVFormatContext *ic = is->ic;
    AVCodecContext *enc;

	if (stream_index < 0 || stream_index >= ic->nb_streams)
        return;
    enc = ic->streams[stream_index]->codec;
        
    switch(enc->codec_type) {
    case CODEC_TYPE_AUDIO:
        packet_queue_abort(&is->audioq);

//		pspAudioSetChannelCallback(0, NULL, NULL);
		close_audio();

		if (is->audio_resample)
		{
			audio_resample_close(is->resample);
		}

        packet_queue_end(&is->audioq);
        break;
    case CODEC_TYPE_VIDEO:
        packet_queue_abort(&is->videoq);

        /* note: we also signal this mutex to make sure we deblock the
           video thread in all cases */
		sceKernelWaitSema(is->pictq_mutex, 1, 0);
		sceKernelSignalSema(is->pictq_mutex, 1);

		sceKernelWaitThreadEnd(is->video_tid, NULL);
		sceKernelDeleteThread(is->video_tid);

        packet_queue_end(&is->videoq);
		scePowerSetClockFrequency(222,222,111);
        break;
#ifdef SUBTITLE
    case CODEC_TYPE_SUBTITLE:
        packet_queue_abort(&is->subtitleq);
        
        /* note: we also signal this mutex to make sure we deblock the
           video thread in all cases */
        SDL_LockMutex(is->subpq_mutex);
        is->subtitle_stream_changed = 1;
    
        SDL_CondSignal(is->subpq_cond);
        SDL_UnlockMutex(is->subpq_mutex);

        SDL_WaitThread(is->subtitle_tid, NULL);

        packet_queue_end(&is->subtitleq);
        break;
#endif
    default:
        break;
    }

    avcodec_close(enc);
    switch(enc->codec_type) {
    case CODEC_TYPE_AUDIO:
        is->audio_st = NULL;
        is->audio_stream = -1;
        break;
    case CODEC_TYPE_VIDEO:
        is->video_st = NULL;
        is->video_stream = -1;
        break;
#ifdef SUBTITLE
    case CODEC_TYPE_SUBTITLE:
        is->subtitle_st = NULL;
        is->subtitle_stream = -1;
        break;
#endif
    default:
        break;
    }
}

void dump_stream_info(AVFormatContext *s)
{
    if (s->track != 0)
        fprintf(stderr, "Track: %d\n", s->track);
    if (s->title[0] != '\0')
        fprintf(stderr, "Title: %s\n", s->title);
    if (s->author[0] != '\0')
        fprintf(stderr, "Author: %s\n", s->author);
    if (s->album[0] != '\0')
        fprintf(stderr, "Album: %s\n", s->album);
    if (s->year != 0)
        fprintf(stderr, "Year: %d\n", s->year);
    if (s->genre[0] != '\0')
        fprintf(stderr, "Genre: %s\n", s->genre);
}

/* since we have only one decoding thread, we can use a global
   variable instead of a thread local variable */
static VideoState *global_video_state;

static int decode_interrupt_cb(void)
{
    return (global_video_state && global_video_state->abort_request);
}

/* this thread gets the stream from the disk or the network */
static int decode_thread(SceSize args, void *argp)
{
    VideoState *is = *(VideoState**)argp;

    AVFormatContext *ic;
    int err, i, ret, video_index, audio_index, use_play;
    AVPacket pkt1, *pkt = &pkt1;
    AVFormatParameters params, *ap = &params;
	char err_msg[2048];

    video_index = -1;
    audio_index = -1;
    is->video_stream = -1;
    is->audio_stream = -1;
#ifdef SUBTITLE
	is->subtitle_stream = -1;
#endif

    global_video_state = is;
    url_set_interrupt_cb(decode_interrupt_cb);

    memset(ap, 0, sizeof(*ap));
    ap->image_format = image_format;
    ap->initial_pause = 1; /* we force a pause when starting an RTSP
                              stream */
    
    err = av_open_input_file(&ic, is->filename, is->iformat, 0, ap);
    if (err < 0) {
        print_error(is->filename, err);
		catched_error = 1;
        ret = -1;
        goto fail;
    }
    is->ic = ic;
#ifdef CONFIG_NETWORK
    use_play = (ic->iformat == &rtsp_demux);
#else
    use_play = 0;
#endif


    if(genpts)
        ic->flags |= AVFMT_FLAG_GENPTS;

    if (!use_play) {
        err = av_find_stream_info(ic);
        if (err < 0) {
            snprintf(err_msg, 2048, "%s: could not find codec parameters\n", is->filename);
			msg_box((const char *)err_msg);
			catched_error = 1;
            ret = -1;
            goto fail;
        }
		ic->pb.eof_reached= 0; //FIXME hack, ffplay maybe shouldnt use url_feof() to test for the end
    }

    /* if seeking requested, we execute it */
    if (start_time != AV_NOPTS_VALUE) {
        int64_t timestamp;

        timestamp = start_time;
        /* add the stream start time */
        if (ic->start_time != AV_NOPTS_VALUE)
            timestamp += ic->start_time;
        ret = av_seek_frame(ic, -1, timestamp, AVSEEK_FLAG_BACKWARD);
        if (ret < 0) {
            fprintf(stderr, "%s: could not seek to position %0.3f\n", 
                    is->filename, (double)timestamp / AV_TIME_BASE);
        }
    }

    /* now we can begin to play (RTSP stream only) */
    av_read_play(ic);

    if (use_play) {
        err = av_find_stream_info(ic);
        if (err < 0) {
            sprintf(err_msg, "%s: could not find codec parameters\n", is->filename);
			msg_box((const char *)err_msg);
			catched_error = 1;
            ret = -1;
            goto fail;
        }
    }

    for(i = 0; i < ic->nb_streams; i++) {
        AVCodecContext *enc = ic->streams[i]->codec;
        switch(enc->codec_type) {
        case CODEC_TYPE_AUDIO:
            if (audio_index < 0 && !audio_disable)
                audio_index = i;
            break;
        case CODEC_TYPE_VIDEO:
            if (video_index < 0 && !video_disable)
                video_index = i;
            break;
        default:
            break;
        }
    }
    if (show_status) {
        dump_format(ic, 0, is->filename, 0);
        dump_stream_info(ic);
    }

    /* open the streams */
    if (audio_index >= 0) {
        stream_component_open(is, audio_index);
    }

    if (video_index >= 0) {
        stream_component_open(is, video_index);
    } else {
        if (!display_disable)
            is->show_audio = 1;
    }

    if (is->video_stream < 0 && is->audio_stream < 0) {
        snprintf(err_msg, 2048, "%s: could not open codecs\n", is->filename);
		msg_box((const char *)err_msg);
		catched_error = 1;
        ret = -1;
        goto fail;
    }

    for(;;) {
        if (is->abort_request)
            break;
#ifdef CONFIG_NETWORK
        if (is->paused != is->last_paused) {
            is->last_paused = is->paused;
            if (is->paused)
                av_read_pause(ic);
            else
                av_read_play(ic);
        }
        if (is->paused && ic->iformat == &rtsp_demux) {
            /* wait 10 ms to avoid trying to get another packet */
            /* XXX: horrible */
			sceKernelDelayThread(10000);
            continue;
        }
#endif
        if (is->seek_req) {
//			pspAudioSetChannelCallback(0, NULL, NULL);
			audio_pause();
            sceKernelWaitSema(is->video_decoder_mutex, 1, 0);
            sceKernelWaitSema(is->audio_decoder_mutex, 1, 0);
#ifdef SUBTITLE
			sceKernelWaitSema(is->subtitle_decoder_mutex, 1, 0);
#endif
            ret = av_seek_frame(is->ic, -1, is->seek_pos, is->seek_flags);

            if (ret < 0) {
                fprintf(stderr, "%s: error while seeking\n", is->ic->filename);
            }else{
                if (is->audio_stream >= 0) {
                    packet_queue_flush(&is->audioq);
                }
#ifdef SUBTITLE
				if (is->subtitle_stream >= 0) {
                    packet_queue_flush(&is->subtitleq);
                }
#endif
                if (is->video_stream >= 0) {
                    packet_queue_flush(&is->videoq);
                    avcodec_flush_buffers(ic->streams[video_index]->codec);
                }
            }
#ifdef SUBTITLE
			sceKernelSignalSema(is->subtitle_decoder_mutex, 1);
#endif
			sceKernelSignalSema(is->audio_decoder_mutex, 1);
			sceKernelSignalSema(is->video_decoder_mutex, 1);

            is->seek_req = 0;
//			pspAudioSetChannelCallback(0, sdl_audio_callback, NULL);
			audio_pause();
        }

        /* if the queue are full, no need to read more */
        if (is->audioq.size > MAX_AUDIOQ_SIZE ||
            is->videoq.size > MAX_VIDEOQ_SIZE || 
#ifdef SUBTITLE
			is->subtitleq.size > MAX_SUBTITLEQ_SIZE || 
#endif
            url_feof(&ic->pb)) {
            /* wait 10 ms */
			if (url_feof(&ic->pb))
			{
				if (is->audioq.size == 0 && is->videoq.size == 0)
					play_finished = 1;
			}
			sceKernelDelayThread(10000);
            continue;
        }
        ret = av_read_frame(ic, pkt);
        if (ret < 0) {
	    	if (url_ferror(&ic->pb) == 0) {
				sceKernelDelayThread(100000);
				continue;
	    	} else
			{
				play_finished = 1;
            	break;
			}
        }
        if (pkt->stream_index == is->audio_stream) {
            packet_queue_put(&is->audioq, pkt);
        } else if (pkt->stream_index == is->video_stream) {
            packet_queue_put(&is->videoq, pkt);
#ifdef SUBTITLE
		} else if (pkt->stream_index == is->subtitle_stream) {
            packet_queue_put(&is->subtitleq, pkt);
#endif
        } else {
            av_free_packet(pkt);
        }
    }
    /* wait until the end */
    while (!is->abort_request) {
		sceKernelDelayThread(100000);
    }

    ret = 0;
 fail:
    /* disable interrupting */
    global_video_state = NULL;

    /* close each stream */
    if (is->audio_stream >= 0)
        stream_component_close(is, is->audio_stream);
    if (is->video_stream >= 0)
        stream_component_close(is, is->video_stream);
#ifdef SUBTITLE
    if (is->subtitle_stream >= 0)
        stream_component_close(is, is->subtitle_stream);
#endif
    if (is->ic) {
        av_close_input_file(is->ic);
        is->ic = NULL; /* safety */
    }
    url_set_interrupt_cb(NULL);

    if (ret != 0) {
    }
    return 0;
}

static VideoState *stream_open(const char *filename, AVInputFormat *iformat)
{
    VideoState *is;

    is = av_mallocz(sizeof(VideoState));
    if (!is)
        return NULL;
    pstrcpy(is->filename, sizeof(is->filename), filename);
    is->iformat = iformat;
//	is->width = 480;
//	is->height = 272;
	is->width = 320;
	is->height = 80;
    is->ytop = 10;
    is->xleft = 80;

    /* start video display */
	is->pictq_mutex = sceKernelCreateSema("", 0, 1, 1, 0);
#ifdef SUBTITLE
    is->subpq_mutex = SDL_CreateMutex();
    is->subpq_cond = SDL_CreateCond();
    
    is->subtitle_decoder_mutex = SDL_CreateMutex();
#endif
    is->audio_decoder_mutex = sceKernelCreateSema("", 0, 1, 1, 0);
    is->video_decoder_mutex = sceKernelCreateSema("", 0, 1, 1, 0);
	is->audio_buffer_mutex = sceKernelCreateSema("", 0, 1, 1, 0);
    /* add the refresh timer to draw the picture */
    schedule_refresh(is, 40);

    is->av_sync_type = av_sync_type;

//    is->parse_tid = sceKernelCreateThread("Decode thread", (SceKernelThreadEntry)decode_thread, 0x09, 0x10000, THREAD_ATTR_VFPU, 0);
    is->parse_tid = sceKernelCreateThread("Decode thread", (SceKernelThreadEntry)decode_thread, 0x13, 0x10000, THREAD_ATTR_VFPU, 0);
	if (is->parse_tid < 0) {
        av_free(is);
        return NULL;
    }
	else
	{
		sceKernelStartThread(is->parse_tid, sizeof(VideoState *), &is);
	}
    return is;
}

static void stream_close(VideoState *is)
{
    VideoPicture *vp;
    int i;
    /* XXX: use a special url_shutdown call to abort parse cleanly */
    is->abort_request = 1;
	sceKernelWaitThreadEnd(is->parse_tid, NULL);
	sceKernelDeleteThread(is->parse_tid);

    /* free all pictures */
    for(i=0;i<VIDEO_PICTURE_QUEUE_SIZE; i++) {
        vp = &is->pictq[i];
        if (vp->bmp) {
#ifdef JONNY_YUV2RGB
			free_64(vp->buffer); // [jonny]
			av_free(vp->bmp);
			vp->buffer = NULL;
#else
			free_overlay_surface(vp->bmp);
#endif
			vp->bmp = NULL;
        }
    }
	sceKernelDeleteSema(is->pictq_mutex);
#ifdef SUBTITLE	
	SDL_DestroyMutex(is->subpq_mutex);
    SDL_DestroyCond(is->subpq_cond);
    SDL_DestroyMutex(is->subtitle_decoder_mutex);
#endif
    sceKernelDeleteSema(is->audio_decoder_mutex);
    sceKernelDeleteSema(is->video_decoder_mutex);
	sceKernelDeleteSema(is->audio_buffer_mutex);
// Jini ...
	av_free(is);
}

void stream_cycle_channel(VideoState *is, int codec_type)
{
    AVFormatContext *ic = is->ic;
    int start_index, stream_index;
    AVStream *st;

    if (codec_type == CODEC_TYPE_VIDEO)
        start_index = is->video_stream;
	else if (codec_type == CODEC_TYPE_AUDIO)
        start_index = is->audio_stream;
#ifdef SUBTITLE
    else
        start_index = is->subtitle_stream;
    if (start_index < (codec_type == CODEC_TYPE_SUBTITLE ? -1 : 0))
#else
	if (start_index < 0)
#endif
        return;
    stream_index = start_index;
    for(;;) {
        if (++stream_index >= is->ic->nb_streams)
		{
#ifdef SUBTITLE
            if (codec_type == CODEC_TYPE_SUBTITLE)
            {
                stream_index = -1;
                goto the_end;
            } else
#endif
            stream_index = 0;
		}
        if (stream_index == start_index)
            return;
        st = ic->streams[stream_index];
        if (st->codec->codec_type == codec_type) {
            /* check that parameters are OK */
            switch(codec_type) {
            case CODEC_TYPE_AUDIO:
                if (st->codec->sample_rate != 0 &&
                    st->codec->channels != 0)
                    goto the_end;
                break;
            case CODEC_TYPE_VIDEO:
#ifdef SUBTITLE
			case CODEC_TYPE_SUBTITLE:
#endif
                goto the_end;
            default:
                break;
            }
        }
    }
 the_end:
    stream_component_close(is, start_index);
    stream_component_open(is, stream_index);
}

void toggle_pause(void)
{
    if (cur_stream)
        stream_pause(cur_stream);
    step = 0;
}

void step_to_next_frame(void)
{
    if (cur_stream) {
        if (cur_stream->paused)
            cur_stream->paused=0;
        cur_stream->video_current_pts = get_video_clock(cur_stream);
    }
    step = 1;
}

void do_exit(void)
{
    if (cur_stream) {
        stream_close(cur_stream);
        cur_stream = NULL;
    }
    if (show_status)
        printf("\n");
}

void toggle_audio_display(void)
{
    if (cur_stream) {
        cur_stream->show_audio = !cur_stream->show_audio;
    }
}

void reset_display_clock(void *priv)
{
	display_clock = 0;
}

void change_cpu_clock(void)
{
	int cpu_freq;

	cpu_freq = scePowerGetCpuClockFrequency();
	if (cpu_freq >= 333)
	{
		scePowerSetClockFrequency(266,266,133);
		snprintf(clock_text, 256, "clock : %d", 266);
	}
	else
	{
		scePowerSetClockFrequency(333,333,166);
		snprintf(clock_text, 256, "clock : %d", 333);
	}

	display_clock = 1;

	set_timer(2000, reset_display_clock, NULL, 0);
}

/* handle an event sent by the GUI */
void event_loop(void)
{
    double incr, pos, frac;
	myevent event;

    for(;;) {
		get_event(&event);
		if (event.value & PSP_CTRL_START)
		{
//			toggle_pause();
			sceKernelDelayThread(1000);
			mode++;
			mode %= 3;
			set_disp_mode(cur_stream, mode);
			clear_screen();
			disp_update();
			clear_screen();
			disp_update();
//			toggle_pause();
		}
		if (event.value & PSP_CTRL_TRIANGLE)
		{
//			toggle_pause();
//			change_cpu_clock();
//			toggle_pause();
			repeat = repeat ? 0 : 1;
		}
		if (event.value & PSP_CTRL_SQUARE)
		{
			if (refresh_timer_id >= 0)
				reset_timer(refresh_timer_id);
//			toggle_pause();
			do_exit();
			return;
		}
		if (event.value & PSP_CTRL_CROSS)
		{
			toggle_pause();
		}
		if (event.value & PSP_CTRL_LEFT)
		{
			if (cur_stream->video_st)
			{
//				toggle_pause();
				incr = -10.0;
				goto do_seek;
			}
			else
			{
			}
		}
		if (event.value & PSP_CTRL_RIGHT)
		{
			if (cur_stream->video_st)
			{
//				toggle_pause();
				incr = 10.0;
				goto do_seek;
			}
			else
			{
			}
		}
		if (event.value & PSP_CTRL_LTRIGGER)
		{
			if (cur_stream->video_st)
			{
//				toggle_pause();
				incr = -300.0;
				goto do_seek;
			}
			else if (cur_stream->audio_st)
			{
//				toggle_pause();
				do_exit();
				if (get_prev_file(MEDIA_MUSIC))
				{
					input_filename = FileName;
					cur_stream = stream_open(input_filename, file_iformat);
				}
				else
					return;
			}
		}
		if (event.value & PSP_CTRL_RTRIGGER)
		{
			if (cur_stream->video_st)
			{
//				toggle_pause();
				incr = 300.0;
				goto do_seek;
			}
			else if (cur_stream->audio_st)
			{
//				toggle_pause();
				do_exit();
				if (get_next_file(MEDIA_MUSIC))
				{
					input_filename = FileName;
					cur_stream = stream_open(input_filename, file_iformat);
				}
				else
					return;
			}
		}
		if (event.value & PSP_CTRL_UP)
		{
			if (cur_stream->video_st)
			{
//				toggle_pause();
				incr = 60.0;
				goto do_seek;
			}
		}
		if (event.value & PSP_CTRL_DOWN)
		{
			if (cur_stream->video_st)
			{
//				toggle_pause();
				incr = -60.0;
do_seek:
        	    if (cur_stream) {
            	    pos = get_master_clock(cur_stream);
                	pos += incr;
                    stream_seek(cur_stream, (int64_t)(pos * AV_TIME_BASE), incr);
    	        }
//				toggle_pause();
			}
		}
		if (need_alloc)
		{
			alloc_picture(cur_stream);
			need_alloc = 0;
		}
/* Jini .. change video refresh using timer 051117 */
//		if (need_timer_check && av_gettime() >= next_refresh_time)
//		{
//			need_timer_check = 0;
//			video_refresh_timer(cur_stream);
//		}
		if (event.value & MY_EVENT_REFRESH)
		{
			video_refresh_timer(event.userdata);
		}
/* Jini .. change video refresh to use timer 051117 */
		if (play_finished)
		{
			play_finished = 0;
			if (cur_stream->audio_st && cur_stream->show_audio)
			{
//				toggle_pause();
				do_exit();
				if (repeat)
				{
					cur_stream = stream_open(input_filename, file_iformat);
				}
				else if (get_next_file(MEDIA_MUSIC))
				{
					input_filename = FileName;
					cur_stream = stream_open(input_filename, file_iformat);
				}
				else
				{
					return;
				}
			}
		}
		if (return_from_sleep)
		{
			int music;
			return_from_sleep = 0;
//			toggle_pause();
			if (cur_stream->audio_st && cur_stream->show_audio)
				music = 1;
			else
				music = 0;
			do_exit();
			sceKernelDelayThread(500000);
			strcpy(FileName, saved_filename);
			input_filename = FileName;
			cur_stream = stream_open(input_filename, file_iformat);
			if (!music)
			{
				sceKernelDelayThread(500000);
				stream_seek(cur_stream, (int64_t)(saved_pos * AV_TIME_BASE), 0);
			}
		}
		if (catched_error)
		{
			catched_error = 0;
			return;
		}
		if (done)
		{
			exit_program();
			return;
		}
		sceKernelDelayThread(10);
    }
}

void file_list_reset()
{
}

int file_list_init(media_type type)
{
  int result;
  int i, len;
  int handle, ret;

  result = 1;

  if (type == MEDIA_VIDEO)
  	snprintf(media_dir_name, MAXPATHLEN, "ms0:/PSP/VIDEO/");
  else
    snprintf(media_dir_name, MAXPATHLEN, "ms0:/PSP/MUSIC/");
  handle = sceIoDopen(media_dir_name);
  if (handle < 0)
  {
	result = 0; /* failed */
  }
  else
  {
    i = 0;
	ret = 1;
    while ( (ret > 0) && i < MAX_MEDIA_FILES ) {
	  ret = sceIoDread(handle, &media_file_list[i]);
	  if (!FIO_SO_ISREG(media_file_list[i].d_stat.st_attr)) /* not a file so skip */
	  {
	    continue;
	  }
	  if (ret > 0)
	    i++;
	}
	media_file_num = i;
	list_end = media_file_num - 1;
  }
  sceIoDclose(handle);
  if (media_file_num == 0)
  {
    result = 0;
  }
  return result;
}

int media_list_init(media_type type)
{
  return file_list_init(type);
}

#define TOP_HEIGHT 0
#define ITEM_HEIGHT 18
#define BOTTOM_HEIGHT 20
#define ITEM_NUM ((SCREEN_HEIGHT - 40) / ITEM_HEIGHT)
#define LINE_PAD 2

void clear_screen()
{
  my_rect area;
  unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);

  area.x = 0;
  area.y = 0;
  area.w = LINESIZE;
  area.h = SCREEN_HEIGHT;
  my_clear_area(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK);
  disp_copy(screen_buffer, LINESIZE, 0, 0, LINESIZE, SCREEN_HEIGHT, 0, 0);
}

void draw_item(int scrn_idx, int file_idx, u8 *text, u8 scrolbar)
{
	int start_x, start_y;
	int text_len;
	my_rect area;
	MY_Surface* textbuf;
	MY_Color fg = {0xFF, 0xFF, 0xFF, 0};

	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);

	area.x = 20;
	area.y = scrn_idx * ITEM_HEIGHT + 20;
	area.w = (SCREEN_WIDTH - 40);
	area.h = ITEM_HEIGHT;

	if (text)
	{  
		start_x = 0;
		text_len = strlen((void *)text);
		start_y = LINE_PAD;
		if (!use_ttf_font)
		{
	    	my_draw_text(screen_buffer, LINESIZE, start_x, start_y, text, text_len, MY_COLOR_WHITE, area);
		}
		else
		{
   			ascii2unicode((u8 *)text, unicode_text);
			textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, fg);
			MY_BlitSurface(screen_buffer, LINESIZE, textbuf, area);
			MY_FreeSurface(textbuf);
		}
	}
}
void draw_item_selected(int scrn_idx, int file_idx, u8 *text, u8 scrolbar)
{
	int start_x, start_y;
	int text_len;
	my_rect area;
	MY_Surface* textbuf;
	MY_Color fg = {0xFF, 0xFF, 0x00, 0};
	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);

	if (text == NULL)
	{
		return;
	}

	area.x = 20;
	area.y = scrn_idx * ITEM_HEIGHT + 20;
	area.w = (SCREEN_WIDTH - 40);
	area.h = ITEM_HEIGHT;

	my_clear_area_alpha(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_LIGHT_GRAY, 200);

	start_x = 0;
	text_len = strlen((void *)text);
	start_y = LINE_PAD;
	if (!use_ttf_font)
	{
		my_draw_text(screen_buffer, LINESIZE, start_x, start_y, text, text_len, MY_COLOR_YELLOW, area);
	}
	else
	{
  		ascii2unicode((u8 *)text, unicode_text);
		textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, fg);
		MY_BlitSurface(screen_buffer, LINESIZE, textbuf, area);
		MY_FreeSurface(textbuf);
	}
}

void draw_media_list(int last_start, int start, int last, int current, int end)
{
  int i;
  int scrn_idx;
  int index;
  int start_y;
  my_rect area;
  unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);

	if (background)
	{
		my_copy_bmp(screen_buffer, LINESIZE, SCREEN_HEIGHT, 0, 0, background);
		area.x = 20;
		area.y = 20;
		area.w = (SCREEN_WIDTH - 40);
		area.h = (SCREEN_HEIGHT - 40);
		my_clear_area_alpha(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK, 126);
	}
	else
	{
		area.x = 0;
		area.y = 0;
		area.w = SCREEN_WIDTH;
		area.h = SCREEN_HEIGHT;
		my_clear_area(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_WHITE);
	}

    index = start;
    scrn_idx = 0;
    has_scrolbar = 0;
    for (i = 0; i < ITEM_NUM; i++)
	{
	  if (index <= end)
	  {
        if (index == current)
		{
		  draw_item_selected(scrn_idx, index, (u8 *)media_file_list[index].d_name, has_scrolbar);
		}
		else
		{
		  draw_item(scrn_idx, index, (u8 *)media_file_list[index].d_name, has_scrolbar);
		}
	  }
	  else
	  {
		draw_item(scrn_idx, 0, NULL, has_scrolbar);
	  }
	  scrn_idx++;
	  index++;
	}
  disp_copy((char *)screen_buffer, LINESIZE, 0, 0, LINESIZE, SCREEN_HEIGHT, 0, 0);
}

void show_info_box(const char *msg)
{
	my_rect area;
	my_doc *doc;
	my_text *p;
	const char *text;
	int len, y;
	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);
	MY_Surface* textbuf;
	MY_Color msg_c = {253, 114, 30, 0};

	if (!msg)
		return;

	if (use_ttf_font)
		TTF_ChangeFontSize(my_font.font, TITLE_PTSIZE);

	area.x = 40;
	area.y = 20;
	area.w = SCREEN_WIDTH - 80;
	area.h = SCREEN_HEIGHT - 40;

	my_clear_area(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK);

	area.x += 2;
  	area.w -= 4;

	len = strlen((const char *)msg);
  	doc = my_str_to_doc((const u8 *)msg, len, area.w);

  	p = doc->head->next;
  	if (doc->linecnt * NORMAL_FONT_HEIGHT < area.h)
  	{
    	area.y += (area.h - doc->linecnt * NORMAL_FONT_HEIGHT) / 2;
    	area.h = doc->linecnt * NORMAL_FONT_HEIGHT;
  	}
  	else
  	{
    	area.y += 2;
    	area.h -= 4;
  	}
  	y = 0;
  	while (p != doc->tail)
  	{
    	text = (const char *)p->text;
		len = p->len;
		while (*text == ' ')
		{
	  		text++;
	  		len--;
		}
		if (len > 0)
		{
			if (!use_ttf_font)
			{
     			my_draw_aligned_text(screen_buffer, LINESIZE, y, (const u8 *)text, len, ALIGN_MIDDLE, MY_COLOR_MSG, area);
			}
			else
			{
				my_rect msg_area;

			   	ascii2unicode((u8 *)text, unicode_text);
				textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, msg_c);
				msg_area.x = area.x + (area.w - textbuf->w) / 2;
				msg_area.y = area.y + y;
				msg_area.w = textbuf->w;
				msg_area.h = textbuf->h;
				MY_BlitSurface(screen_buffer, LINESIZE, textbuf, msg_area);
				MY_FreeSurface(textbuf);
			}
		}
    	y += NORMAL_FONT_HEIGHT;
    	if (y > (area.h - NORMAL_FONT_HEIGHT))
      		break;
    	p = p->next;
  	}
  	my_doc_release(doc);

	disp_copy((char *)screen_buffer, LINESIZE, 0, 0, LINESIZE, SCREEN_HEIGHT, 0, 0); 
	disp_update();
	if (use_ttf_font)
		TTF_ChangeFontSize(my_font.font, DEFAULT_PTSIZE);
}

void msg_box(const char *msg)
{
	myevent event;

	my_rect area;
	my_doc *doc;
	my_text *p;
	const char *text;
	int len, y;
	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);
	MY_Surface* textbuf;
	MY_Color msg_c = {0xFF, 0xFF, 0xFF, 0};
	MY_Color button_c = {0xFF, 0x00, 0x00, 0};

	if (!msg)
		return;

	if (background)
	{
		my_copy_bmp(screen_buffer, LINESIZE, SCREEN_HEIGHT, 0, 0, background);
	}
	else
	{
		area.x = 0;
		area.y = 0;
		area.w = SCREEN_WIDTH;
		area.h = SCREEN_HEIGHT;
		my_clear_area(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK);
	}
	area.x = 40;
	area.y = 20;
	area.w = SCREEN_WIDTH - 80;
	area.h = SCREEN_HEIGHT - 40;

	my_clear_area_alpha(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK, 200);
	my_draw_rect(screen_buffer, LINESIZE, SCREEN_HEIGHT, area.x, area.y, area.w, area.h, MY_COLOR_LIGHT_GRAY);

	area.x += 2;
  	area.w -= 4;

	len = strlen((const char *)msg);
  	doc = my_str_to_doc((const u8 *)msg, len, area.w);

  	p = doc->head->next;
  	if (doc->linecnt * NORMAL_FONT_HEIGHT < area.h)
  	{
    	area.y += (area.h - doc->linecnt * NORMAL_FONT_HEIGHT) / 2;
    	area.h = doc->linecnt * NORMAL_FONT_HEIGHT;
  	}
  	else
  	{
    	area.y += 2;
    	area.h -= 4;
  	}
  	y = 0;
  	while (p != doc->tail)
  	{
    	text = (const char *)p->text;
		len = p->len;
		while (*text == ' ')
		{
	  		text++;
	  		len--;
		}
		if (len > 0)
		{
			if (!use_ttf_font)
			{
     			my_draw_aligned_text(screen_buffer, LINESIZE, y, (const u8 *)text, len, ALIGN_MIDDLE, MY_COLOR_MSG, area);
			}
			else
			{
				my_rect msg_area;

			   	ascii2unicode((u8 *)text, unicode_text);
				textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, msg_c);
				msg_area.x = area.x + (area.w - textbuf->w) / 2;
				msg_area.y = area.y + y;
				msg_area.w = textbuf->w;
				msg_area.h = textbuf->h;
				MY_BlitSurface(screen_buffer, LINESIZE, textbuf, msg_area);
				MY_FreeSurface(textbuf);
			}
		}
    	y += NORMAL_FONT_HEIGHT;
    	if (y > (area.h - NORMAL_FONT_HEIGHT))
      		break;
    	p = p->next;
  	}
  	my_doc_release(doc);
	area.x = 40;
	area.y = 20;
	area.w = SCREEN_WIDTH - 80;
	area.h = SCREEN_HEIGHT - 40;

	text = (const char *)mytxt_get_string(MY_TXT_PRESS_CROSS, language);
	if (!use_ttf_font)
	{
		len = strlen(text);
		my_draw_aligned_text(screen_buffer, LINESIZE, area.h - 20, (const u8 *)text, len, ALIGN_MIDDLE, MY_COLOR_RED, area);
	}
	else
	{
	   	ascii2unicode((u8 *)text, unicode_text);
		textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, button_c);
		area.x = area.x + (area.w - textbuf->w) / 2;
		area.w = textbuf->w;
		area.y = SCREEN_HEIGHT - 60;
		area.h = textbuf->h;
		MY_BlitSurface(screen_buffer, LINESIZE, textbuf, area);
		MY_FreeSurface(textbuf);
	}

	disp_copy((char *)screen_buffer, LINESIZE, 0, 0, LINESIZE, SCREEN_HEIGHT, 0, 0); 
	disp_update();

	while (1)
	{
		if (done)
			break;
		get_event(&event);
		if (event.value & PSP_CTRL_CROSS)
		{
			return;
		}
	}
	if (done)
	{
		exit_program();
	}
}

void init_background()
{
	char buf[MAXPATHLEN];
	getcwd(buf, MAXPATHLEN);
	pstrcat(buf, MAXPATHLEN, "/background.jpg");
	background = read_jpeg_file(buf);
}

void reset_background()
{
	free_img(background);
	background = NULL;
}

void greeting()
{
	myevent event;

	my_rect area;
	char *intro_txt;
	int len;
	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);

	area.x = 0;
	area.y = 0;
	area.w = SCREEN_WIDTH;
	area.h = SCREEN_HEIGHT;
	my_clear_area(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK);

	if (background)
	{
		my_copy_bmp(screen_buffer, LINESIZE, SCREEN_HEIGHT, 0, 0, background);
	}

	area.x = 50;
	area.y = 50;
	area.w = (SCREEN_WIDTH - 100);
	area.h = (SCREEN_HEIGHT - 100);

	my_clear_area_alpha(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK, 200);

	intro_txt = (char *)mytxt_get_string(MY_TXT_INTRO, language);  
	if (!use_ttf_font)
	{
		len = strlen(intro_txt);
		my_draw_multi_text(screen_buffer, LINESIZE, 0, 0, (const u8 *)intro_txt, len, 1, MY_COLOR_WHITE, area);
	}
	else
	{
		ascii2unicode((u8 *)intro_txt, unicode_text);
		my_ttf_draw_text(screen_buffer, LINESIZE, 0, 0, unicode_text, area, my_font.font, MY_COLOR_WHITE);
	}

	intro_txt = (char *)mytxt_get_string(MY_TXT_KEY_HELP, language);
	if (!use_ttf_font)
	{
		len = strlen(intro_txt);
		my_draw_multi_text(screen_buffer, LINESIZE, 0, 90, (const u8 *)intro_txt, len, 1, MY_COLOR_WHITE, area);
	}
	else
	{
		ascii2unicode((u8 *)intro_txt, unicode_text);
		my_ttf_draw_text(screen_buffer, LINESIZE, 0, 90, unicode_text, area, my_font.font, MY_COLOR_WHITE);
	}
	disp_copy((char *)screen_buffer, LINESIZE, 0, 0, LINESIZE, SCREEN_HEIGHT, 0, 0); 
	disp_update();

	while (1)
	{
		if (done)
			break;
		get_event(&event);
		if (event.value & PSP_CTRL_SELECT)
		{
			return;
		}
		if (event.value & PSP_CTRL_START)
		{
			done = 1;
		}
		sceKernelDelayThread(50000);
	}
	if (done)
	{
		exit_program();
	}
}

void draw_select_type(media_type current)
{
	int y = 0;
	const char * txt;
	int len;
	unsigned int selected_c, unselected_c;
	my_rect area;
	unsigned int *screen_buffer = (unsigned int *)pgGetVramAddr(0,0);
	MY_Surface* textbuf;
	MY_Color red = {0xFF, 0x00, 0x00, 0};
	MY_Color gray = {0xB4, 0xB4, 0xB4, 0};
	MY_Color selected;
	MY_Color unselected;
	if (current == MEDIA_VIDEO)
	{
		selected_c = MY_COLOR_RED;
		unselected_c = MY_COLOR_LIGHT_GRAY;
		selected = red;
		unselected = gray;
	}
	else
	{
		selected_c = MY_COLOR_LIGHT_GRAY;
		unselected_c = MY_COLOR_RED;
		selected = gray;
		unselected = red;
	}

	if (background)
	{
		my_copy_bmp(screen_buffer, LINESIZE, SCREEN_HEIGHT, 0, 0, background);
	}
	else
	{
		area.x = 0;
		area.y = 0;
		area.w = LINESIZE;
		area.h = SCREEN_HEIGHT;
		my_clear_area(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK);
	}

	area.x = 50;
	area.y = 50;
	area.w = (SCREEN_WIDTH - 100);
	area.h = (SCREEN_HEIGHT - 100);

	my_clear_area_alpha(screen_buffer, LINESIZE, SCREEN_HEIGHT, area, MY_COLOR_BLACK, 200);

 	txt = (const char *)mytxt_get_string(MY_TXT_VIDEO, language);
	if (!use_ttf_font)
	{
		area.x = 0;
		area.y = 0;
		area.w = LINESIZE;
		area.h = SCREEN_HEIGHT;

	 	y = (SCREEN_HEIGHT - NORMAL_FONT_HEIGHT * 2) / 2;
	 	len = strlen((const char *)txt);
	 	my_draw_aligned_text(screen_buffer, LINESIZE, y, (const u8 *)txt, len, ALIGN_MIDDLE, selected_c, area);
	}
	else
	{
	   	ascii2unicode((u8 *)txt, unicode_text);
		textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, selected);
		area.x = (SCREEN_WIDTH - textbuf->w) / 2;
		area.y = (SCREEN_HEIGHT - textbuf->h*2) / 2;
		area.w = textbuf->w;
		area.h = textbuf->h;
		MY_BlitSurface(screen_buffer, LINESIZE, textbuf, area);
		MY_FreeSurface(textbuf);
	}
 	txt = (const char *)mytxt_get_string(MY_TXT_MUSIC, language);
	if (!use_ttf_font)
	{
	 	y += 20;
	 	len = strlen((const char *)txt);
 		my_draw_aligned_text(screen_buffer, LINESIZE, y, (const u8 *)txt, len, ALIGN_MIDDLE, unselected_c, area);
	}
	else
	{
	   	ascii2unicode((u8 *)txt, unicode_text);
		textbuf = TTF_RenderUNICODE_Solid(my_font.font, unicode_text, unselected);
		area.x = (SCREEN_WIDTH - textbuf->w) / 2;
		area.y += textbuf->h;
		area.w = textbuf->w;
		area.h = textbuf->h;
		MY_BlitSurface(screen_buffer, LINESIZE, textbuf, area);
		MY_FreeSurface(textbuf);
	}
 	disp_copy((char *)screen_buffer, LINESIZE, 0, 0, LINESIZE, SCREEN_HEIGHT, 0, 0);
 	disp_update();
}

media_type select_type()
{
	myevent event;

	media_type type = MEDIA_VIDEO; // default
	if (use_ttf_font)
		TTF_ChangeFontSize(my_font.font, TITLE_PTSIZE);
	draw_select_type(type);
	while (1)
	{
		if (done)
			break;
		get_event(&event);
		if (event.value & PSP_CTRL_UP)
		{
			if (type == MEDIA_VIDEO)
			{
			}
			else
			{
				type = MEDIA_VIDEO;
				draw_select_type(type);
			}
		}
		if (event.value & PSP_CTRL_DOWN)
		{
			if (type == MEDIA_MUSIC)
			{
			}
			else
			{
				type = MEDIA_MUSIC;
				draw_select_type(type);
			}
		}
		if (event.value & PSP_CTRL_CIRCLE)
		{
			if (use_ttf_font)
				TTF_ChangeFontSize(my_font.font, DEFAULT_PTSIZE);
			return type;
		}
		sceKernelDelayThread(50000);
	}
	if (done)
	{
		exit_program();
	}
	if (use_ttf_font)
		TTF_ChangeFontSize(my_font.font, DEFAULT_PTSIZE);
	return type;	  
}

char * select_file(media_type type)
{
	myevent event;
	char *filename = NULL;

    last_start_idx = -1;
    last_media_idx = 0;
    start_idx = 0;
    media_idx = 0;
    list_end = -1;
	if (media_list_init(type))
	{
		draw_media_list(last_start_idx, start_idx, last_media_idx, media_idx, list_end);
		disp_update();
	}
	else
	{
		if (type == MEDIA_VIDEO)
		{
			const char *txt;
			txt = (const char *)mytxt_get_string(MY_TXT_NO_VIDEO_FILE, language);
			msg_box(txt);
		}
		else
		{
			const char *txt;
			txt = (const char *)mytxt_get_string(MY_TXT_NO_MUSIC_FILE, language);
			msg_box(txt);
		}
		return filename;
	}
	while (1)
	{
		if (done)
			break;

		get_event(&event);
		if (event.value & PSP_CTRL_UP)
		{
			if ((media_idx - 1) < 0)
			{
			  last_media_idx = media_idx;
			  media_idx = list_end;
			  last_start_idx = start_idx;
		      if (list_end >= ITEM_NUM)
			  {
			    start_idx = list_end - ITEM_NUM + 1;
			  }
			  else
			  {
			    start_idx = 0;
			  }
			}
			else if ((media_idx - 1) < start_idx)
			{
			  last_media_idx = media_idx;
		      media_idx--;
			  last_start_idx = start_idx;
		      start_idx = media_idx;
		    }
			else
			{
			  last_media_idx = media_idx;
			  media_idx--;
		      last_start_idx = start_idx;
			}
			draw_media_list(last_start_idx, start_idx, last_media_idx, media_idx, list_end);
			disp_update();
		}
		if (event.value & PSP_CTRL_DOWN)
		{
		 	if ((media_idx + 1) > list_end)
		 	{			  
		 	  last_media_idx = media_idx;
		 	  media_idx = 0;
		 	  last_start_idx = start_idx;
		 	  start_idx = 0;
		 	}
		 	else if ((media_idx + 1) >= start_idx + ITEM_NUM)
		 	{
		 	  last_media_idx = media_idx;
		       media_idx++;
		 	  last_start_idx = start_idx;
		       start_idx = media_idx - ITEM_NUM + 1;
		     }
		 	else
		 	{
		 	  last_media_idx = media_idx;
		 	  media_idx++;
		       last_start_idx = start_idx;
		 	}
		 	draw_media_list(last_start_idx, start_idx, last_media_idx, media_idx, list_end);
		 	disp_update();
		}
		if (event.value & PSP_CTRL_CIRCLE)
		{
			/* õ ϸ copyϰ ÷  */
			if (type == MEDIA_VIDEO)
				strcpy(FileName, "ms0:/PSP/VIDEO/");
			else
				strcpy(FileName, "ms0:/PSP/MUSIC/");
			pstrcat(FileName, MAXPATHLEN, media_file_list[media_idx].d_name);
			filename = FileName;
			break;
		}
		if (event.value & PSP_CTRL_TRIANGLE)
		{
			return NULL;
		}
		sceKernelDelayThread(50000);
	}
	if (done)
	{
		exit_program();
	}
	return filename;
}

void set_disp_mode(VideoState *is, screen_mode mode)
{
	double ratio;
	int width, height;

	width = is->video_st->codec->width;
	height = is->video_st->codec->height;
	if (height > 272)
	{
		mode = FIT_MODE;
	}
	switch(mode)
	{
		case FIT_MODE:
			ratio = (double)272.0f / (double)height;
			disp_w = width * ratio;
			disp_h = 272;
			break;
		case FULL_SCREEN:
			disp_w = 480;
			disp_h = 272;
			break;
		case NORMAL_MODE:
		default:
			disp_w = width;
			disp_h = height;
			break;
	}
	if (disp_w > 480)
		disp_w = 480;
}

int get_prev_file(media_type type)
{
	if ((media_idx - 1) < 0)
	{
	  last_media_idx = media_idx;
	  media_idx = list_end;
	  last_start_idx = start_idx;
       if (list_end >= ITEM_NUM)
	  {
	    start_idx = list_end - ITEM_NUM + 1;
	  }
	  else
	  {
	    start_idx = 0;
	  }
	}
	else if ((media_idx - 1) < start_idx)
	{
	  last_media_idx = media_idx;
       media_idx--;
	  last_start_idx = start_idx;
       start_idx = media_idx;
     }
	else
	{
	  last_media_idx = media_idx;
	  media_idx--;
       last_start_idx = start_idx;
	}
	if (last_media_idx == media_idx)
		return 0;
	if (type == MEDIA_VIDEO)
		strcpy(FileName, "ms0:/PSP/VIDEO/");
	else
		strcpy(FileName, "ms0:/PSP/MUSIC/");
	pstrcat(FileName, MAXPATHLEN, media_file_list[media_idx].d_name);
	return 1;
}

int get_next_file(media_type type)
{
	if ((media_idx + 1) > list_end)
	{			  
	  last_media_idx = media_idx;
	  media_idx = 0;
	  last_start_idx = start_idx;
	  start_idx = 0;
	}
	else if ((media_idx + 1) >= start_idx + ITEM_NUM)
	{
	  last_media_idx = media_idx;
        media_idx++;
	  last_start_idx = start_idx;
        start_idx = media_idx - ITEM_NUM + 1;
      }
	else
	{
	  last_media_idx = media_idx;
	  media_idx++;
        last_start_idx = start_idx;
	}
	if (last_media_idx == media_idx)
		return 0;
	if (type == MEDIA_VIDEO)
		strcpy(FileName, "ms0:/PSP/VIDEO/");
	else
		strcpy(FileName, "ms0:/PSP/MUSIC/");
	pstrcat(FileName, MAXPATHLEN, media_file_list[media_idx].d_name);
	return 1;
}
void ascii2unicode(u8 *src, u16* dest)
{
	unsigned int data[2048];
	int len;
	int index;

	len = ufEuc_kr2Unicode(data, src);
	for (index = 0; index < len; index++) {
		dest[index] = data[index];
	}
	dest[index] = 0;
}

int init_ttf_font(char *filename, font_info *font, int ptsize)
{
	int result = 0;

	if (filename)
	{
		strcpy(font->filename, filename);
		if (TTF_Init() >= 0)
		{
			font->ptsize = ptsize;
			font->renderstyle = TTF_STYLE_NORMAL;
			font->rendersolid = 1;
			font->font = TTF_OpenFont(filename, font->ptsize);
			if (font->font)
			{
				TTF_SetFontStyle(font->font, font->renderstyle);
				result = 1;
			}
		}
	}
	return result;
}

void close_ttf_font(font_info *font)
{
	TTF_CloseFont(font->font);
}

void exit_program(void)
{
	do_exit();
	done = 1;
}

int debug_main(SceSize args, void* argp)
{
    /* register all codecs, demux and protocols */
    av_register_all();

	input_filename = "ms0:/test.ogg";
	while (!done)
	{
	   	cur_stream = stream_open(input_filename, file_iformat);
		event_loop();
	}
	return 0;
}

/* Called from the main */
//int main(int argc, char **argv)
int player_main(SceSize args, void* argp)
{
	media_type type;
	const char *txt;

//	pspDebugScreenInit();
//	SetupCallbacks();
//	pspDebugInstallErrorHandler(MyExceptionHandler);

	if (use_ttf_font)
	{
		if (!init_ttf_font(DEFAULT_FONTNAME, &my_font, DEFAULT_PTSIZE))
		{
			if (!init_font())
			{
				printf("Failed to initialize font!!!");
				return 1;
			}
			use_ttf_font = 0;
			ttf_failed = 1;
		}
	}
	else
	{
		if (!init_font())
		{
			printf("Failed to initialize font!!!");
			return 1;
		}
	}

	init_background();
	init_event_handler();
	pspAudioInit();

    /* register all codecs, demux and protocols */
    av_register_all();

	pgInit();
#ifndef JONNY_YUV2RGB
	init_palettes();
#endif
	while(1)
	{
		greeting();
		if (done)
			break;
		type = select_type();
		input_filename = select_file(type);
		if (input_filename == NULL)
			continue;
		clear_screen();
		disp_update();
		clear_screen();
		disp_update();

		txt = (const char *)mytxt_get_string(MY_TXT_LOADING, language);
		show_info_box(txt);

    	cur_stream = stream_open(input_filename, file_iformat);
		event_loop();
	}
	pgShutDown();
	pspAudioEnd();
	reset_background();
	close_event_handler();
	if (use_ttf_font && !ttf_failed)
	{
		close_ttf_font(&my_font);
	}
	else
	{
		close_font();
	}
    return 0;
}

int main(void)
{
    // Kernel mode thread
	pspDebugScreenInit();
	SetupCallbacks();
	pspDebugInstallErrorHandler(MyExceptionHandler);
//#ifdef JONNY_YUV2RGB
//	p_me_struct = me_struct_init();               // [jonny]
//	me_startproc((u32) me_function, p_me_struct); // [jonny]
//#endif
    me_idct_p_malloc(); // [jonny]
	me_struct_init();   // [jonny]

    // create user thread, tweek stack size here if necessary
//    SceUID thid = sceKernelCreateThread("Player Thread", player_main,
//            0x11, // default priority
//            0x20000, // stack size (256KB is regular default)
//            PSP_THREAD_ATTR_USER, NULL);
    SceUID thid = sceKernelCreateThread("Player Thread", player_main,
            0x13, // default priority
            0x20000, // stack size (256KB is regular default)
            PSP_THREAD_ATTR_USER, NULL);

    // start user thread, then wait for it to do everything else
    sceKernelStartThread(thid, 0, NULL);
    sceKernelWaitThreadEnd(thid, NULL);

    sceKernelExitGame();
    return 0;
}